/*------------------------------------------------------------------------*\
                               SCOTT.c                                   
                                                                         
    This file contains the functions called by OTTER when compiled
    with SCOTT defined. They set up the data structures specific to
    SCOTT and call the requisite FINDER.
  
    SCOTT uses the FINDER subsystem to determine approximately maximal
    consistent sets (maxisets) of the evolving clause space.  These
    maxisets are used to implement a guidance strategy.

    Suppose M1, M2, M3,...,Mn are maximally consistent set on our
    clause space.  Then (by completeness) every proof must contain at
    least one clause from each of the complements M1', M2',...,Mn'.
    There may, of course, be some necessary infernce that will belong
    to the initerection M^1M^2M^3...^Mn, but since proofs tend to be
    short and maxisets plenetiful, they will be in the minorirty.  In
    order to increase the probability that these necessary inferences
    are made, we restrict our selection to each complement in turn,
    and then the intersection.

    The code for this is spread across SCOTT.c, Sglob.h, Sdef.h,
    Spartition.c, and Sclause_space.c (that is, all files beginning
    with an "S").

    Main contact points from OTTER to SCOTT:

    1. semantic selection() selects best SOS clause based on semantic
       criterion.

    2. add_semantic_wt() incorporates new clauses into the existing
       collection of maxisets.

    4. scott_insert(), scott_delete() utilities add/remove clauses
       to/form the partition record and the scanning order.

    Main contact points from SCOTT to FINDER

    1. model() attempts to extend the current theory with a given
    clause.  

    2. declare_otter_symbols() generates function declarations for the
       signature of the given problem.

\*------------------------------------------------------------------------*/

#define IN_SCOTT

#include "SCOTT.h"
#include <stdio.h>
#include <stdarg.h>

#define MAX_HISTORY 100000
static char History[MAX_HISTORY];

static int givens_avoided;
static int givens_avoided_by[MAXISET_ARRAY];
static int last_num_updates[MAXISET_ARRAY];

/*----------------------------------------------------------------------*/

int semantic_wt(struct clause *cl) 
  /*------------------------------------------------------------------
   * Called by add_semantic_wt(), model_clause(), test_clause(), and
   * decendents of check_partition().
   * 
   * Returns the semantic weight of the given clause. Clauses that
   * belong to more than one maxisets have their weight altered,
   * according to the paramter TRUE_WEIGHT
   * -----------------------------------------------------------------*/
{
  if (cl->num_sets < 2) return (cl->length);
  else {
    char str[5];
    int weight = 0;
    strcpy(str,"");
    
    // a very chunky round function
    sprintf(str,"%.0f",
	    (cl->num_sets -1)
	    *tParms[TRUE_WEIGHT].val
	    /(float)num_maxisets);
    
    sscanf(str,"%d",&weight);
    
    return (cl->length + weight);
  }
}/*--semantic_wt() ---------------------------------------------------*/

void add_semantic_wt(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by OTTER fcn pre_process().
   * 
   * Fattens new clause according to the number of sets (maxisets)
   * that contain the clause. Also sets semantic information about the
   * clause.
   * 
   * If in the course of this call the search time limit changes, the
   * global variable time_limit_has_changed is set.
   * -----------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];
  int maxiset;

  SCOTT_CLOCK_START(SEM_GUIDE_TIME);
  if (cl->first_lit) {
    if(!make_finder_clause(finder_cl, cl)) {
      SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
      return;
    }
    for (maxiset = 1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset]) {
	if (model(finder_cl,maxiset,0) ){ 
	  cl->in_set[maxiset] = true;
	  cl->num_sets++;
	}
      }
    }    
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }
  SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
  
}/*--add_semantic_wt() -----------------------------------------------*/

void retest_failed_clauses(int maxiset)
  /*------------------------------------------------------------------
   * Called by add_semantic_wt().  
   * 
   * The model of given maxiset has just been updated so all the
   * clauses external to that maxiset are retested in the new model.
   * Because these clause have already failed once, a new search is
   * still quite likely to fail with the given resource limits, hence
   * we only test the clause in the new model.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next;
  struct clause *cl;

  message(3,"Retesting failed clauses");

  cp = failed_clauses[maxiset].first_cp;
  while (cp  && cp->c) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[maxiset]) 
      abort_scott("Function retest_failed_clauses()"
		  "called with clause already true in model");
    
    if (test_clause(cl,maxiset)) {

      if (cl->container == Sos
	  || cl->container == First){
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
	sos_part_mv_to_set(cl,maxiset);
      }
      else usable_part_mv_to_set(cl,maxiset);
      delete_cl_ext(cl,&failed_clauses[maxiset]);
      tStats[SUCCESSFUL_RETESTS]++;
    }

    if (!Flags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }

} /*--retest_failed_clauses() ----------------------------------------*/

void remodel_failed_sos_clauses(int maxiset)
  /*------------------------------------------------------------------
   * Called by semantic_selection()
   * 
   * Our bound on search time has just changed, so we remodel SOS
   * clauses to improve the quality of guidance.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next;
  struct clause *cl;

  cp = failed_clauses[maxiset].first_cp;
  while (cp && cp->c) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[maxiset]) 
      abort_scott("Function remodel_failed_clauses() called"
		  "with clause already true in model");

    if (cl->container == Sos) { // only remodel SOS
      if (model_clause(cl,maxiset)) {
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
	sos_part_mv_to_set(cl,maxiset);
	delete_cl_ext(cl,&failed_clauses[maxiset]);
	tStats[SUCCESSFUL_RETESTS]++;
      }
    }
    if (!Flags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }
  
  
} /*--remodel_failed_sos_clauses() ----------------------------------------*/


void check_failed_clauses()
  /*------------------------------------------------------------------
   * Called by semantic_selection().  
   * 
   * Performs two checks in the following order:
   * 
   * A) If our time limit has changed between calls to this function,
   * extension of each maxiset by each of its failed SOS clauses is
   * attempted.
   *
   * B) If any maxiset has a new model (witnessing its consistency)
   * each of its failed clauses (all lists) are retested.
   * -----------------------------------------------------------------*/
{
  static int prev_time_limit = 0;
  float est_cost;
  int time_left, maxiset, last_time;

  last_time = 0;

  if ( 0 
       && num_maxisets > 0
      && prev_time_limit 
      && get_time_limit() > prev_time_limit) {
    
    last_time = run_time();
    //    reset_optimal();

    //first do we want to do this??
    est_cost = Stats[SOS_SIZE]
      * est_cost_per_clause()
      * num_maxisets
      * get_time_limit()
      / (float)(prev_time_limit);
    
    if (Parms[MAX_SECONDS].val > 0)
      time_left = (Parms[MAX_SECONDS].val*1000) - run_time();
    else time_left = MAX_INT;
    
    if (est_cost*fParms[COST_MARGIN].val < time_left) {
      est_cost /= 1000;
      message(3,"Remodelling failed clauses (estimated cost %.1f secs)", 
	      est_cost);
      for (maxiset = 1; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) remodel_failed_sos_clauses(maxiset);
      if (!Flags[SEMANTIC_GUIDANCE].val) return;
      print_partition("after failed clauses remodelled");   
    }
  }
  prev_time_limit = get_time_limit();
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]
	&& last_num_updates[maxiset]
	&& get_updates(maxiset) > last_num_updates[maxiset]){
      retest_failed_clauses(maxiset);
      if (!Flags[SEMANTIC_GUIDANCE].val) return;
    }
    last_num_updates[maxiset] = get_updates(maxiset);
  }

  if (0 
      && last_time 
      && num_maxisets > 0) {// just in case 
    set_cost_per_clause(
			(run_time()-last_time)
			/(float) num_maxisets
			/(float) (Stats[SOS_SIZE]+Stats[USABLE_SIZE])
			);
  }
}/*check_failed_clauses()---------------------------------------------*/


struct clause *semantic_selection(struct list *lptr)
  /*------------------------------------------------------------------
   * Called by find_given_cl().
   * 
   * Select the next given clause according to the following semantic
   * criterion.
   * 
   * 1. Select the first clause on the "input_first" list if one
   *    exists.  
   * 2. Select the first clause on the SOS list if it has
   *    fewer than SEM_QUEUE_LIMIT clauses.  
   * 3. Move appropriate clause to the "input_first" list if we are
   *    performing a semantic queue, and select the first clause form
   *    that list.
   * 4. Select the first, lightest SOS clause that does not belong to
   *    the current set to skip.
   * -----------------------------------------------------------------*/
{
  static boolean first = false;
  static boolean semantic = false;
  static boolean standard = false;
  static int count = 0;
  int last_num_attempts = 0;
  int set_to_skip = 0;
  static float last_time = 0;

  SCOTT_DEBUG("before semantic_selection()");
  
  SCOTT_CLOCK_START(SEM_GUIDE_TIME);

  if (Flags[SEMANTIC_GUIDANCE].val) {
    set_cost_of_given(run_time() - last_time);
    set_time_saved(givens_avoided);
  }

  if (First->first_cl) {
    // catch remaining semantic queue clauses
    if (!first) {
      record_event("\n%dm->f...",num_maxisets);
      first = true;
      standard = semantic = false;
      count = 0;
    }
    count++;
    tStats[GIVENS_FROM_FIRST]++;
    last_time = run_time();
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return First->first_cl;
  }
  
  if (!lptr->first_cl) {
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return NULL; // henceforth assume list is non-empty
  }

  // revise semantic information
  if (Flags[SEMANTIC_GUIDANCE].val) {
    
    check_failed_clauses();
    if (discard_bad_maxisets()){
      record_event("(%d)...",count);
      count = 0;
    }    
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
 
    if (develop_semantics) {
      SCOTT_CLOCK_START(SEM_DEV_TIME);
      
      last_num_attempts = tStats[ATTEMPTS];
   
      SCOTT_DEBUG("before cover_new_clauses()");
      if (execute_strategy(cover_new_clauses)){
	SCOTT_DEBUG("before refine_choice()");
	execute_strategy(refine_choice);
	SCOTT_DEBUG("after refine_choice()");
      }

      if (tStats[ATTEMPTS] > last_num_attempts) {
	first = standard = semantic = false;
	record_event(" (%d)",count);
      }
      SCOTT_CLOCK_STOP(SEM_DEV_TIME);        
    }
    else if (num_maxisets < 2) {
      disable_semantics("Semantic guidance too expensive to continue");
    }

    SCOTT_CLOCK_START(SEM_GUIDE_TIME);
    check_failed_clauses();
    set_to_skip = next_set_to_skip();
  }

  set_sem_queue(lptr);

  // catch any new input first clauses
  if (First->first_cl) {
    if (!first) {
      record_event("\n%dm->b...",num_maxisets);
      first = true;    
      standard = semantic = false;
      count = 0;
    }
    count++;
    tStats[GIVENS_FROM_FIRST]++;
    last_time = run_time();
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return First->first_cl;
  }
  if (first) record_event(" (%d)",count);
  first = false;    

  // catch first semantic queue clause
  if (First->first_cl) {
    if (!first) {
      record_event("\n%dm->b...",num_maxisets);
      first = true;    
      standard = semantic = false;
      count = 0;
    }
    count++;
    tStats[GIVENS_FROM_FIRST]++;
    last_time = run_time();
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return First->first_cl;
  }
  if (first) record_event(" (%d)",count);
  first = false;    

  // standard selection
  if (!Flags[SEMANTIC_GUIDANCE].val || 
      num_maxisets == 0 || set_to_skip == -1) {
    if (!standard) {
      record_event("\n%dm->l...\n",num_maxisets);
      standard = true;
      first = semantic = false;
    }
    count++;
    last_time = run_time();
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return (find_lightest_cl(lptr));
  }
  standard = false;
  
  // else semantic selection
  
  if (!semantic) {
    record_event("\n%dm->s...",num_maxisets);
    semantic = true;
    first =  false;
    count = 0;
  }
  count++;
  last_time = run_time();
  return (find_best_sem_cl(lptr,set_to_skip));
  
}/*--semantic_selection()---------------------------------------------*/




struct clause *find_best_sem_cl(struct list *lptr, int set_to_skip)
  /*------------------------------------------------------------------
   * Called by semantic_selection() iff SCOTT has been defined.
   * 
   * Returns the "best" semantic clause on the list.  That is, the
   * lightest clause in the given list, belonging to the complement of
   * the current set_to_skip, and such that its selection will
   * minimise the variance in the distribution of complements on the
   * usable list.
   * -----------------------------------------------------------------*/
{
  struct clause *cl, *best = NULL;

  message(3,"skip maxiset number %d",set_to_skip);

  if (set_to_skip == 0) {
    // get lightest clause belonging to the most maxisets
    best = lptr->first_cl;
    for (cl = best->next_cl; cl; cl = cl->next_cl){
      if (cl->num_sets > best->num_sets ||
	  cl->pick_weight < best->pick_weight ) 
	best = cl;
    }
  }
  else if (Flags[SMOOTH_CMP_DIST].val) {

    float best_variance = 0;
    float next_variance = 0;

    for (cl = lptr->first_cl; cl; cl = cl->next_cl) {
      // skip clauses in the set_to_skip
      if (cl->in_set[set_to_skip]) {
	// clauses that we would have otherwise chosen
	givens_avoided_by[set_to_skip]++;
	givens_avoided++;
	continue;
      }

      if (!best) {
	best = cl;
	best_variance = variance(best);
	continue;
      }

      // should this include equality?
      // the SG strategy is stronger than SF so no
      if (cl->pick_weight > best->pick_weight ) continue;

      if (cl->num_sets < best->num_sets) continue;

      next_variance = variance(cl);    
      if (abs(next_variance-best_variance)/best_variance > 
	  fParms[SMOOTHING_RATIO].val) {
	givens_avoided_by[set_to_skip]++;
	givens_avoided++;
	continue;
      }      
      best = cl;
      best_variance = next_variance;
    }
  }
  else {
    // get first lightest clause not in the skip_theory
    for (cl = lptr->first_cl; cl; cl = cl->next_cl){
      if (cl->in_set[set_to_skip]) {
	// clauses that we would have otherwise chosen
	givens_avoided_by[set_to_skip]++;
	givens_avoided++;
	continue;
      }
      if (!best) {
	best = cl;
	continue;
      }
      if (cl->pick_weight < best->pick_weight ) best = cl;
    }
  }

  if (best) tStats[SEMANTIC_SELECTIONS]++;

  SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
  return best;

}/*--find_best_cl() --------------------------------------------------*/

void extend_base(int key_maxiset)
  /*------------------------------------------------------------------
   * Called if execute_strategy() and discard_bad_maxisets().
   * -----------------------------------------------------------------*/
{
  struct clause *new_base_cl = NULL;

  message(3,"");
  message(3, "----> Attempting to extend the semantic base");

  new_base_cl = search_list_1(Usable,key_maxiset);
  
  if (!new_base_cl) {
    message(3,"No suitable on USABLE list...searching SOS");
    new_base_cl = search_list_1(Sos,key_maxiset);
  }
  
  if (!new_base_cl) {
    message(3,"No suitable clause on SOS...weakening conditions");
    new_base_cl = search_list_2(Usable,key_maxiset);
  }
  
  if (!new_base_cl) {
    message(3,"No suitable clause on USABLE list...searching SOS");
    new_base_cl = search_list_2(Sos,key_maxiset);
  }

  if (!new_base_cl) {
    message(3,"No suitable clause on SOS...weakening conditions");
    new_base_cl = search_list_3(Usable,key_maxiset);
  }
  if (!new_base_cl) {
    message(3,"No suitable clause on USABLE list...searching SOS");
    new_base_cl = search_list_3(Sos,key_maxiset);
  }

  if (!new_base_cl) {  
    record_event(STRENGTHEN_BASE_FAILURE);
    tStats[BASE_EXTENSION_FAILS]++;
    message(2,"<---- Could not find a suitable clause to add to the semantic base");
    shuffle_ext_list(&new_base_clauses);
  }
  else {
    char in_set[MAXISET_ARRAY+3];

    in_set_to_str(in_set,new_base_cl);
    message(3,"<---- Extending the semantic base with clause %d %s", 
	    new_base_cl->id, in_set);
    //  print_clause(stdout,*new_base_cl);
    new_base_cl->in_base = neither;// to avoid cycles
    prepend_cl_ext(new_base_cl,&new_base_clauses);
  }

  message(3,"");

} /*--extend_base() -------------------------------------------------*/



boolean execute_strategy(boolean (*strategy)())
  /*------------------------------------------------------------------*\
   * Called by find_best_cl().  
   *
   * Attempts to form new maxisets under the given strategy.  Persists
   * so long as the expected improvement in guidance is greater than
   * the expected cost of an attempt.  Discards resultant maxisets
   * that fail the retention test.  Give up if the number of
   * consecutive failures exceeds the parameter MAX_FAILS.
   * 
   * Returns prematurely if FINDER is disabled at any point.
   * -----------------------------------------------------------------*/
{
  int new_maxiset, fails;
  int last_time, this_time, last_time_limit;
  struct clause_dptr *cp;
  
  if (!Flags[SEMANTIC_GUIDANCE].val) return false;
  
  fails = 0; // count consecutive failiures
  
  while (fails < tParms[MAX_FAILS].val && (*strategy)(TEST) && persist()) { 

    last_time = run_time();
    last_time_limit = get_time_limit();

    (*strategy)(RECORD_EVENT);

    /* find the next available maxiset (there must be at least one --
       we have a temporary position set aside)*/
    for (new_maxiset = 1; 
	 new_maxiset < last_set_index && active[new_maxiset]; 
	 new_maxiset++);
    if (active[new_maxiset]) abort_scott("all indices are active");

    num_maxisets++;
    active[new_maxiset] = true;

    if (tParms[MIN_MAXISET_ATTEMPTS].val < tStats[ATTEMPTS]) 
      set_prob_of_success(tStats[KEPT]/(float)(tStats[ATTEMPTS]));
    tStats[ATTEMPTS]++;
    
    start_scan(new_maxiset);  if (!Flags[SEMANTIC_GUIDANCE].val) return false;
    scan_base(new_maxiset);   if (!Flags[SEMANTIC_GUIDANCE].val) return false;
    scan_sos(new_maxiset);    if (!Flags[SEMANTIC_GUIDANCE].val) return false;
    scan_first(new_maxiset);  if (!Flags[SEMANTIC_GUIDANCE].val) return false;
    scan_usable(new_maxiset); if (!Flags[SEMANTIC_GUIDANCE].val) return false;
    
    update_partition(new_maxiset); // this must go before next function
    SCOTT_DEBUG("after update_partition()");

    /*
    if (new_max_wt) cull_oversized_clauses(); // assume partition is ok
    new_max_wt = false;
    SCOTT_DEBUG("after cull_oversized_clauses()");
    */

    if (!(*strategy)(FAILS) && keep(new_maxiset)) {
      
      // move new base clauses to the semantic base
      for (cp = new_base_clauses.first_cp; cp; cp = cp->next) {
	if (cp->c->in_set[new_maxiset]) {
	  cp->c->in_base = true;
	  prepend_cl_ext(cp->c,&semantic_base);
	  tStats[SEM_BASE_SIZE]++;
	  tStats[BASE_CL_ADDED]++;
	  if (cp->c->container == Sos){
	    message(3,"Moving new base clause %d from SOS to INPUT_FIRST",
		    cp->c->id);
	    rem_from_list(cp->c);
	    append_cl(First,cp->c);
	    SCOTT_DEBUG("after moving SOS to INPUT_FIRST");
	  }
	}
	else cp->c->in_base = false; // reset from "neither"
      }
      // discard the existing new_base_clauses
      for (cp = new_base_clauses.first_cp;cp;
	   cp = new_base_clauses.first_cp)
	delete_cl_ext(cp->c,&new_base_clauses);
      
      tStats[KEPT]++;    
      (*strategy)(RECORD_KEPT);
      print_partition("after a new maxiset added");
      SCOTT_DEBUG("after new maxiset kept");
      
      fails = 0; 
      discard_subsets(new_maxiset);
      if (get_time_limit() > last_time_limit){
	reset_prob_of_success();  
	//	reset_optimal();
      }
      SCOTT_DEBUG("after subsets discarded");
    }
    else {
      if (get_time_limit() > last_time_limit) {
	message(2,"Resetting search time limit to %d", 
		last_time_limit);
	set_time_limit(last_time_limit);
      } 
      check_for_model(new_maxiset); // may not return
      if (!Flags[SEMANTIC_GUIDANCE].val) return false;
      reset_partition(new_maxiset);
      message(2,"<==== Discarding new maxiset %d",new_maxiset);
      message(2,"");
      discard(new_maxiset);    
      shuffle_ext_list(&sos_scan[0]);
      shuffle_ext_list(&sos_scan[1]);
      shuffle_ext_list(&sos_scan[2]);
      SCOTT_DEBUG("after new maxiset discarded");
      fails++;
    }    
    this_time = run_time();
    set_cost_per_clause((this_time - last_time)/(float)
			(Stats[SOS_SIZE]+Stats[USABLE_SIZE]));
    last_time = this_time;
  }

  if (fails == tParms[MAX_FAILS].val) return false;
  else return true;

}/*--execute_strategy()-----------------------------------------------*/

void update_partition(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy() after scan.  Updates the partition
   * data structures to include the new maxiset.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    sos_part_insert(cl,new_maxiset);

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    sos_part_insert(cl,new_maxiset);
  
  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    usable_part_insert(cl,new_maxiset);

}/*--update_partition()-----------------------------------------------*/

void reset_partition(int bad_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy() after scan.  Removes the bad_maxiset
   * from the partition data structures.  NB must be called before
   * clause membership data is reset.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  // reset the SOS
  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    sos_part_delete(cl,bad_maxiset);

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    sos_part_delete(cl,bad_maxiset);
  
  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    usable_part_delete(cl,bad_maxiset);

}/*--reset_partition()-----------------------------------------------*/


boolean extend_conjecture(int new_maxiset)
  /*------------------------------------------------------------------*\
   * Called by keep().  A maxiset has just been generated that covers
   * the whole clause space except for one clause
   * Move this clause to the conjecture list. Always append to the conjecture list.
   * Otherwise we may make it too strong .. excluding new base clauses.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  for (cl = Usable->first_cl; cl; cl = cl->next_cl) {
    if (!cl->parents // input clause
	&& cl->in_base != neither // not already a conjecture
	&& !cl->in_set[new_maxiset]) {
      message(2,"Extending conjecture list with USABLE clause %d",cl->id);
      append_cl_ext(cl,&conjecture);
      cl->in_base = neither;
      return true;
    }
  }
  /*
  for (cl = Sos->first_cl; cl; cl = cl->next_cl) {
    if (!cl->parents // input clause
	&& cl->in_base != neither // not already a conjecture
	&& !cl->in_set[new_maxiset]) {
      message(2,"Extending conjecture list with SOS clause %d.",cl->id);
      message(2,"Moving clause %d to the INPUT_FIRST list",cl->id);
      append_cl_ext(cl,&conjecture);
      rem_from_list(cl);
      append_cl(First,cl);
      cl->in_base = neither;
      return true;;
    }
  }
  */
  return false;

}/*extend_conjecture() -----------------------------------------------*/


boolean keep(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy().
   * 
   * Retension test for newly generated maxiset.  

   * Note: discarding this maxiset will reorder the whole scan so that
   * new_maxiset clauses come last, decreasing the likelihood that a
   * duplicate maxiset set will be formed next time.
   * -----------------------------------------------------------------*/
{
  float rating = 0;
  int superset = 0;
  int size_on_sos, size_on_usable;

  size_on_sos = get_size_on_sos(new_maxiset);
  size_on_usable = get_size_on_usable(new_maxiset);

  // check if we have covered the whole SOS (hence offers no distinction)
  if (Stats[SOS_SIZE] > tParms[SEM_QUEUE_LIMIT].val
      && size_on_sos >= Stats[SOS_SIZE]) {

    message(2,"Maxiset %d covers whole SOS",new_maxiset);

    if (Stats[SOS_SIZE] > tParms[SEM_QUEUE_LIMIT].val
	&& (size_on_sos + size_on_usable + 1)
	== (Stats[SOS_SIZE] + Stats[USABLE_SIZE])) {
      message(2,"Maxiset %d covers all but one clause in the clause space",
	      new_maxiset);
      //extend_conjecture(new_maxiset);// bad function!!
    }

    if (develop_semantics) {// invert the base
      struct clause *cl;
      char in_set[MAXISET_ARRAY+3];
      
      for (cl=Usable->first_cl; cl; cl=cl->next_cl)
	if(!cl->in_base) {
	  in_set_to_str(in_set,cl);
	  append_cl_ext(cl,&new_base_clauses);
	  cl->in_base = neither;
	  message(3,"<----Extending the semantic base with clause %d %s", 
		  cl->id, in_set);	
	}
      extend_base(new_maxiset);  
    }
    record_event(COVERS_SOS);
    tStats[SOS_COVERS]++;
    return false;
  }

  // check if a subset of any existing theory (and hence not maximal)
  if ((superset = superset_of(new_maxiset))) {

    message(2,"Maxiset %d is a subset of maxiset %d", 
	    new_maxiset, superset);
    
    if (develop_semantics) extend_base(superset);
    record_event(FOR_SUB_MAXISET);
    tStats[FOR_SUBS]++;
    return false;
  }

  // keep maxisets that give us queue clauses
  set_sem_queue(Sos);
  if (First->first_cl) return true;

  // Ensure new maxiset decreases the size of the choice space.  This
  // also ensures the maxiset has a non-empty SOS.
  if ((rating = rate_maxiset(new_maxiset)
       + tParms[MIN_REFINEMENT].val) > 0 ) {
    
    message(2,"Maxiset %d does not refine our choice: rating = %0.1f",
	    new_maxiset,rating);
    
    if (develop_semantics) extend_base(new_maxiset);
    record_event(FAILS_TO_REFINE);
    tStats[REFINE_FAILS]++;
    return false;
  }
  
  return true;
    
}/*--keep()-----------------------------------------------------------*/

boolean discard_bad_maxisets()
  /*------------------------------------------------------------------
   * Called by semantic_selection().
   * 
   * Discards any maxisets in the current rose that no longer provide
   * adequate guidance.  Properties required are the same as for the
   * function keep().
   * -----------------------------------------------------------------*/
{
  int maxiset;  
  boolean bad_sets = false;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) return false;
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset] && !keep(maxiset)) {
      bad_sets = true;
      message(2,"<==== Discarding existing maxiset %d", maxiset);
      reset_partition(maxiset);
      discard(maxiset);
      tStats[CULLED]++;
    }
  }
  
  return bad_sets;

} /*--discard_bad_maxisets() -----------------------------------------*/

void print_in_set(struct external_list *elist)
 /*------------------------------------------------------------------
   * Called by shuffle_ext_list().
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp;
  char in_set[MAXISET_ARRAY+3];
  
  for (cp = elist->first_cp;cp && cp->c; cp = cp->next){
    in_set_to_str(in_set,cp->c);
    printf("%3d %s\n", cp->c->id, in_set);
  }
  fflush(stdout);

}/* print_in_set()----------------------------------------------------*/


void shuffle_ext_list(struct external_list *elist)
  /*------------------------------------------------------------------
   * Called by keep() and refine_choice(). Attempts to shuffle the
   * clauses in the given list so the the in_set fields of adjacent
   * clauses are different.
   * -----------------------------------------------------------------*/
{      
  struct clause_dptr *cp;
  struct clause *prev_cl, *next_cl;
  struct external_list new_elist, failed;
  int count_before = 0;
  int count_mid = 0;
  int count_after = 0;

  /*
  printf("\nexternal list before shuffle\n");
  print_in_set(elist);
  */

  // debugging code
  for (cp = elist->first_cp;cp && cp->c; cp = cp->next)
    count_before++;
  
  new_elist.first_cp = new_elist.last_cp = NULL;
  failed.first_cp = failed.last_cp = NULL;
  
  cp = elist->first_cp;
  
  if (cp && cp->c) { // there must be at least one clause

    count_mid++;    
    prev_cl = cp->c;
    delete_cl_ext(prev_cl,elist);
    append_cl_ext(prev_cl,&new_elist);
    cp = elist->first_cp;
    while (cp && cp->c) {
      
      count_mid++;
      next_cl = cp->c;
      delete_cl_ext(next_cl,elist);
      
      if (!in_set_identical(next_cl,prev_cl)) {
	append_cl_ext(next_cl,&new_elist);
	prev_cl = next_cl;
      }
      else if (!in_set_identical(next_cl,new_elist.first_cp->c))
	prepend_cl_ext(next_cl,&new_elist);
      else if (count_mid % 3 ||
	       count_mid % 5) // shuffle identical guys
      	append_cl_ext(next_cl,&failed);
      else 
      	prepend_cl_ext(next_cl,&failed);

      cp = elist->first_cp;
    }
    
    // debugging code
    if (count_before != count_mid) abort_scott("bad shuffle");
    /*
      printf("\nexternal list mid shuffle\n");
      print_in_set(&new_elist);
      printf("\n");
      print_in_set(&failed);
    */

    // copy the shuffled clauses across to the elist
    // which should be empty by now.
    cp =  new_elist.first_cp;
    while (cp && cp->c) {
      count_after++;
      delete_cl_ext(cp->c,&new_elist);
      append_cl_ext(cp->c,elist);
      cp =  new_elist.first_cp;
    }

    // append matching and prepend non-matching failed clause to the
    // elist this should preserve the shuffe at least at the beginning
    cp = failed.first_cp;
    while (cp && cp->c) {
      count_after++;
      next_cl = cp->c;
      delete_cl_ext(next_cl,&failed);
      
      if (!in_set_identical(next_cl,elist->first_cp->c))
	prepend_cl_ext(next_cl,elist);
      else
	append_cl_ext(next_cl,elist);
      
      cp = failed.first_cp;
    }
  }

  // debugging code
  if (count_after != count_mid) abort_scott("bad shuffle");
  /*
    printf("\nexternal list after shuffle\n");
    print_in_set(elist);
  */
}/*shuffle_ext_list()-------------------------------------------------*/


void scan_new_base_clauses(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by start_scan() 
   *
   * Scan the clauses pointed to by new_base_clauses.  Remove clauses
   * from the list that we fail to model or all those following
   * MAX_NEW_BASE_CL clause passed.
   * -----------------------------------------------------------------*/
{
  int count;
  struct clause_dptr *cp, *next;
  struct clause *cl;

  count = 0;
  cp = new_base_clauses.first_cp;
  while (cp) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    // clause already scanned
    if (cl->in_set[new_maxiset]) 
      delete_cl_ext(cl,&new_base_clauses);
    else {
      count++;
      if (count > tParms[MAX_NEW_BASE_CL].val 
	  || !model_clause(cl, new_maxiset)){
	cl->in_base = false; // reset from neither
	delete_cl_ext(cl,&new_base_clauses);
      }
    }

    cp = next;
  }

} /*scan_new_base_clauses()-------------------------------------------*/

void scan_conjecture(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by start_scan() and auto_set_sos(). 
   * 
   * Scan the clauses in external list "conjecture", removing those
   * that cannot bwe modelled.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next_cp;
  struct clause *cl;

  // model each conjecture clause
  cp = conjecture.first_cp; 
  while (cp){
    next_cp = cp->next;
    cl = cp->c;

    if (model_clause(cl,new_maxiset)) {
      cl->in_base = neither;
    }
    else {
      delete_cl_ext(cl,&conjecture);
      cl->in_base = false;
    }

    cp = next_cp;
  }
}

boolean start_scan(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy()
   *
   * Some newly generated clauses may belong to none of our maxisets.
   * Since we have no semantic information about these clauses yet, we
   * don't know how to treat them.  Hence we place these clauses near
   * the beginning of the scan so that the next maxiset will contain
   * at least one of them.
   * 
   * We start our scan by modelling each clause that was placed in the
   * initial "input_first" list, that is, the designated conjecture
   * clauses.  We continue our scan (if possible) with a single SOS
   * clause that belongs to no maxisets.  Returns true iff such a
   * clause is modelled.
   *
   * (We have no semantic information about these clauses yet, and
   * hence, we don't know how to treat them, so we place these clauses
   * near the beginning of the scan so that the next maxiset will
   * contain at least one of them.)
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp;
  boolean new_max_vars = false;
  boolean modelled = false;
  int last_time_limit, variables;
  
  // allow for extending search time here
  last_time_limit = get_time_limit();
  set_time_limit(tParms[MAX_SEARCH_TIME].val);
  
  // model each conjecture clause
  if (conjecture.first_cp) 
    scan_conjecture(new_maxiset);

  if (new_base_clauses.first_cp) 
    scan_new_base_clauses(new_maxiset);

  // run thru' sos_scan[0] until at least one is modelled
  cp = sos_scan[0].first_cp;
  while (cp) {
    
    // skip clauses already in the null maxiset
    if (!cp->c->in_set[0] && !cp->c->in_base) {

      // model the psuedo null clause
      if (model_clause(cp->c,new_maxiset)) {
	/* moved to scan_sos()
	  delete_cl_ext(cp->c,&sos_scan[0]);
	  append_cl_ext(cp->c,&sos_scan[1]); // deprecate similar scans
	*/
	modelled = true;
	break;
      }	
      if (!Flags[SEMANTIC_GUIDANCE].val) return false;	
      
      // a genuine null set clause has been generated
      cp->c->in_set[0] = true;
      cp->c->pick_weight = semantic_wt(cp->c);
      null_set_size++;
      message(3,"Could not model clause %d with the conjecture",cp->c->id);

      // Note: extension with this clause will still be attempted in scan_sos()

      /*------------------------------------------------------------------
	Clauses actually contradicting the conjecture will normally be
	picked up by the unit conflick check; if we fail to model the
	clause, then it is probably because the clause has too many
	distinct variables.
	-----------------------------------------------------------------*/

      variables = distinct_vars(cp->c);
      if (variables < Parms[MAX_DISTINCT_VARS].val &&
	  variables > variable_percentile(90)) {
	output_variable_distn();
	Parms[MAX_DISTINCT_VARS].val = variables - 1 ;
	message(3,"Resetting distinct variable limit to %d",
		Parms[MAX_DISTINCT_VARS].val);
	fflush(stdout);
	new_max_vars = true;
      }
    }
    cp = cp->next;
  }

  // if extending time limit was not needed, reset to old value
  if (get_max_time_in(new_maxiset) < last_time_limit){
    set_time_limit(last_time_limit);
  }

  return new_max_vars;

}/*--start_scan()-----------------------------------------------------*/

void scan_base(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy()
   * 
   * We continue our scan with clauses from the semantic base.  All
   * these clauses are on either the USABLE or INPUT_FIRST list, and
   * hence will not be effected by clause selection.  Note that, since
   * all our scans have the semantic base clauses, our maxisets will
   * tend to have a common core. This places a strong restriction on the
   * number of maxisets that we will have to consider.
   *
   * At various points we may choose to prepend new base clauses to
   * our list.  This may cause some later base clause to be omited.
   * There in_base feild is set to false so that they may rejoin the
   * base at some later point.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next_cp;
  struct external_list  base_failures;
  static int prev_limit = 0;
  int current_limit;

  base_failures.first_cp = NULL;
  base_failures.last_cp = NULL;

  // has time limit changed since last call?
  current_limit = get_time_limit();
  if (current_limit != prev_limit) {
    reset_base_markers();
    prev_limit = current_limit;
  }

  // allow for a slightly higher search time limit
  if (current_limit*fParms[TOLERANCE].val
      <= tParms[MAX_SEARCH_TIME].val){
    set_time_limit(current_limit*fParms[TOLERANCE].val);
  }
  else {
    set_time_limit(tParms[MAX_SEARCH_TIME].val);
  }

  cp = semantic_base.first_cp;
  while (cp) {
    next_cp = cp->next; // incase cp is deleted

    if (cp->c->in_set[new_maxiset]) { 
      // may be new base clause, this cuts out a copy
      delete_cl_ext(cp->c,&semantic_base);      
    }
    else if (!model_clause(cp->c,new_maxiset)) {
      message(3,"Skipping clause %d from the semantic base", cp->c->id);
      delete_cl_ext(cp->c,&semantic_base);
      prepend_cl_ext(cp->c,&base_failures);
      tStats[BASE_CL_SKIPPED]++;
    }
    else cp->c->in_base = true;

    if (!Flags[SEMANTIC_GUIDANCE].val) return;
    cp = next_cp;
  }
  
  // second try at base clauses that failed
  cp = base_failures.first_cp;
  while (cp) {
    next_cp = cp->next; // before cp is deleted
    delete_cl_ext(cp->c,&base_failures);  // garbage collection

    if (model_clause(cp->c,new_maxiset)) {

      message(3,"Skipped clause %d is still covered", cp->c->id);
      // note clause is hard to model so place last in base
      append_cl_ext(cp->c,&semantic_base);
      tStats[BASE_CL_SKIPPED]--;
      cp->c->in_base = true;

    }
    if (!Flags[SEMANTIC_GUIDANCE].val) return;  
    
    cp = next_cp;
  }

  fflush(stdout);

  // if extended time limit was not needed, reset to old value
  if (get_max_time_in(new_maxiset) < current_limit){
    set_time_limit(current_limit);
  }

}/*--scan_base() -----------------------------------------------------*/

void reset_base_markers()
   /*------------------------------------------------------------------
   * Called by scan_base(). Resets all in_base fields to false
   * prior to a scan_base() which resets them appropriately
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    if (cl->in_base == true) cl->in_base = false;

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    if (cl->in_base == true) cl->in_base = false;
  
  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    if (cl->in_base == true) cl->in_base = false;

}/*reset_base_markers()-----------------------------------------------*/

void scan_sos(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy()
   * 
   * Completes the new maxiset by scanning the SOS in an order
   * determined by the previous attempt at a model.  This is
   * essentially cell order, or the increasing order of the number of
   * maxisets a clause belongs to.
   * 
   * While performing the scan the order of the next scan altered so
   * that it will be significantly different from this one.
   * 
   * Clauses that are omitted from the new maxiset in the course of
   * the scan are tested against the final model, just in case this
   * model fortuitously bears witness to its consistency.
   * -----------------------------------------------------------------*/
{
  int cell, scanned = 0;
  struct clause *cl;
  struct clause_dptr *cp, *next;
  struct external_list new_sos_scan[MAXISET_ARRAY];

  // init new_sos_scan[]
  for (cell = 0; cell <= last_set_index; cell++) {
    new_sos_scan[cell].first_cp = NULL;
    new_sos_scan[cell].last_cp = NULL;
  }

  for (cell = 0 ; cell <= num_maxisets; cell++){  
    cp = sos_scan[cell].first_cp; 
    while (cp) {
      scanned++;
      cl = cp->c;
      delete_cl_ext(cl,&sos_scan[cell]);// pop off the top element

      if (cl->in_set[new_maxiset]) {      
	// already scanned by start_scan() or scan_base(), so copy to
	// the NEW sos_scan[] and alternate to deprecate similar scans
	if (scanned % 2)  append_cl_ext(cl,&new_sos_scan[cell+1]);
	else prepend_cl_ext(cl,&new_sos_scan[cell+1]);
      }
      else if (model_clause(cl,new_maxiset)) {
	append_cl_ext(cl,&new_sos_scan[cell+1]); // deprecate similar scans
      }
      else {
	if (!Flags[SEMANTIC_GUIDANCE].val) return;
	// alternate to deprecate similar scans 
	if (scanned % 2)  append_cl_ext(cl,&new_sos_scan[cell]);
	else prepend_cl_ext(cl,&new_sos_scan[cell]);

	// prepend so that new failures occur first
	prepend_cl_ext(cl,&failed_clauses[new_maxiset]); 
      }
      cp = sos_scan[cell].first_cp;
    }
  } 

  // note all the ptrs in sos_scan[] have been removed
  for (cell = 0; cell <= last_set_index; cell++) 
    sos_scan[cell] = new_sos_scan[cell];
  
  // retest failed clauses
  cp = failed_clauses[new_maxiset].first_cp;
  while (cp) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[new_maxiset]) 
      abort_scott("Retesting failed clause already true in model");
    
    if (test_clause(cl,new_maxiset)) {

      if (cl->container == Sos) {
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
      }

      // this list should be quite short
      delete_cl_ext(cl,&failed_clauses[new_maxiset]);
      tStats[SUCCESSFUL_RETESTS]++;
    }

    if (!Flags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }

}/*--scan_sos() ------------------------------------------------------*/

void scan_usable(int new_maxiset)
   /*------------------------------------------------------------------
   * Called by execute_strategy().
   * 
   * Extends maxiset with the clauses from the Usable list.  This
   * ensures partition data is correct and that selection of base
   * extensions is optimal.  This may be expensive since the USABLE
   * list will eventually contain many had to model clauses, so after
   * a while we resort to simply testing.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  
  if (model_usable()) { // this should be a parameter
    for (cl= Usable->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) model_clause(cl,new_maxiset);
      if (!Flags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  else {
    for (cl= Usable->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) test_clause(cl,new_maxiset);
      if (!Flags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  
}/*--scan_usable() ------------------------------------------------------*/

void scan_first(int new_maxiset)
   /*------------------------------------------------------------------
   * Called by execute_strategy().
   * 
   * Extends maxiset with the clauses from the INPUT FIRST list.  Note
   * that these will soon be moved to the USABLE LIST, and may be hard
   * to model clauses.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  
  // test the INPUT FIRTS list so partition data is correct
  if (model_usable()) { // this should be a parameter
    for (cl= First->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) model_clause(cl,new_maxiset);
      if (!Flags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  for (cl= First->first_cl; cl; cl = cl->next_cl) {
    // clauses in the semantic base have already been scanned
    if (!cl->in_set[new_maxiset]) test_clause(cl,new_maxiset);
    if (!Flags[SEMANTIC_GUIDANCE].val) return;
  }
  
}/*--scan_usable() ------------------------------------------------------*/


boolean model_clause(struct clause *cl, int maxiset)
  /*------------------------------------------------------------------
   * Called by start_scan(), scan_sos() and scan_base().  
   * 
   * Attempts to extend maxiset with the given clause.  Returns TRUE
   * iff the extension was sucessful.  Returns FALSE if semantics are
   * disabled or if the clause is not modelled
   * -----------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];

  if (!cl->first_lit) return false;  // can't model empty clause!

  if (!make_finder_clause(finder_cl, cl)) {
    return false;
  }

  if (cl->container &&      // cl has been subsummed
      cl->in_set[maxiset])  // there's probably a bug somewhere
    abort_scott("model_clause() gets a clause (%d) that is "
		"already true in the new maxiset", cl->id);

  if (model(finder_cl, maxiset, false)) {
    cl->in_set[maxiset] = true;
    if (cl->in_set[0])  null_set_size--;
    cl->in_set[0] = false;
    cl->num_sets++;
    cl->pick_weight = semantic_wt(cl);
    message(4,"%s %%passed",finder_cl);
  }
  
  message(4,"%s %%failed",finder_cl);
  
  cl->pick_weight = semantic_wt(cl);

  return cl->in_set[maxiset];
  
}/*--model_clause() --------------------------------------------------*/

boolean test_clause(struct clause *cl, int maxiset)
  /*------------------------------------------------------------------
   * Called by scan_sos(), scan_usable() and retest_failed_clauses().  
   * 
   * Tests the given clause in the current model of the given maxiset.
   * If successful, the clause is added to that maxiset and the fcn
   * returns TRUE.  Otherwise, returns FALSE.
   * -----------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];

  if (!cl->first_lit) return false;  // can't model empty clause!

  if (!make_finder_clause(finder_cl, cl)) {
    return false;
  }

  if (cl->in_set[maxiset]) 
    abort_scott("test_clause() gets a clause (%d) that is "
		"already true in the new maxiset", cl->id);

  if (model(finder_cl, maxiset, true)) {  
    cl->in_set[maxiset] = true;
    if (cl->in_set[0]) null_set_size--;
    cl->in_set[0] = false;
    cl->num_sets++;
    message(4,"%s %%passed",finder_cl);
  } 
  message(4,"%s %%failed",finder_cl);

  cl->pick_weight = semantic_wt(cl);

  return cl->in_set[maxiset];
  
}/*--test_clause() ---------------------------------------------------*/



boolean verify_model_in_list(int maxiset, struct list *lptr, char list_name[])
  /*------------------------------------------------------------------
   * Called by verify_model().
   * 
   * Testing each clause in the given list against the model
   * associated with the given maxiset.
   * 
   * Program exits with an error if any clause fails in the model --
   * we should do better here.
   * 
   * This function may not return;
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  char finder_cl[CLAUSE_STRING_ARRAY];

  for (cl = lptr->first_cl; cl ; cl = cl->next_cl) {
    if (!cl->first_lit) { 
      abort_scott(0,"clause space contains an empty clause -- this is a bug");
    }
    if (!make_finder_clause(finder_cl, cl)) {
      message(0,"clause is too large to model");
      return false;
    }
    if (!model(finder_cl, maxiset, true)) {
      message(0,"%s clause fails in propossed model", list_name);
      return false;
    }
  }
  return true;
    
}/*--verify_model_in_list()-------------------------------------------*/

void verify_model(int maxiset)
  /*------------------------------------------------------------------
   * Called by check_for_model()
   * 
   * Our records indicate that the given maxiset covers the whole
   * clause space.  Here we verify that the clause space is infact
   * satisfiable, by testing each clause in each list against the
   * associated model.
   * 
   * Program exist with an error if model fails -- we should do better
   * here.
   * 
   * This function does not return;
   * -----------------------------------------------------------------*/
{

  record_event(VERIFY_MODEL);

  message(0,"Clause space appears to be consistent.\n"
	  "----> Verifying model");

  if (verify_model_in_list(maxiset,Usable,"usable") &&
      verify_model_in_list(maxiset,Sos,"sos") &&
      verify_model_in_list(maxiset,First,"input_first") &&
      verify_model_in_list(maxiset,Demodulators,"demodulator"))
    message(0,"<---- Great SCOTT: Model checks out!!!");
  else
    disable_semantics("<---- proposed model fails!!!");
  
}/*--verify_model()----------------------------------------------------*/

void discard_subsets(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy().  Discards any theories that are
   * sos_subsets of the new_maxiset.
   * -----------------------------------------------------------------*/
{
  int old_maxiset;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val){
    return;
  }

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] && 
	old_maxiset !=  new_maxiset &&
	//subset(old_maxiset,new_maxiset)) { // for real subsets
      sos_subset(old_maxiset,new_maxiset)) {

      message(2,"Maxiset %d is a subset of new maxiset" 
	      "\n<==== Discarding maxiset %d",
	      old_maxiset, old_maxiset);

      record_event(BACK_SUB_MAXISET);

      reset_partition(old_maxiset);
      discard(old_maxiset);

      tStats[BACK_SUBS]++;
      if (old_maxiset == constraint_set) 
	constraint_set = new_maxiset;
    }
  }
}/*--discard_subsets() -----------------------------------------------*/

void discard(int bad_maxiset)
  /*------------------------------------------------------------------
   * Called by discard_bad_maxisets(), 
   *           discard_subsets(),
   *           execute_strategy(), 
   *       and cull_maxisets().
   * 
   * Removes the "bad_maxiset" from the scan, resets the weights of
   * any clause on the SOS true in it, and make its index avaliable
   * for a new maxiset.  Does not change the partition data
   * structures.  These are managed by update_partition() and
   * reset_partition().
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  struct clause_dptr *cp, *next;
  int num_models, scanned;

  fflush(stdout);

  // reset the external clause space lists
  cp = failed_clauses[bad_maxiset].first_cp;
  while (cp) {
    delete_cl_ext(cp->c,&failed_clauses[bad_maxiset]);
    cp = failed_clauses[bad_maxiset].first_cp;
  }

  scanned = 0;
  for (num_models = 1 ; num_models <= num_maxisets; num_models++){  
    cp = sos_scan[num_models].first_cp; 
    while (cp) {
      next = cp->next;
      scanned++;
      // this deletion peels of the top element
      if (cp->c->in_set[bad_maxiset]){
	delete_cl_ext(cp->c,&sos_scan[num_models]);
	// deprecates similar scans
	if (scanned % 3) prepend_cl_ext(cp->c,&sos_scan[num_models-1]);
	else append_cl_ext(cp->c,&sos_scan[num_models-1]);
      }
      cp = next;
    }
  } 

  // make old maxiset position available for next maxiset.
  num_maxisets--;
  active[bad_maxiset] = false;
  if (bad_maxiset == last_queue_theory) last_queue_theory = 0;
  init_sizes(bad_maxiset);
  init_theory(bad_maxiset);

  // reset the DEMODULATORS (all conjecture clauses)
  for (cl=Demodulators->first_cl; cl; cl=cl->next_cl){
      cl->in_set[bad_maxiset] = false;
  }
  
  // reset the SOS
  for (cl=Sos->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }

  // reset the INPUT FIRST list
  for (cl=First->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }
  
  // reset the USABLE list
  for (cl=Usable->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }

}/*--discard()--------------------------------------------------------*/

void disable_semantics(char *event) 
  /*------------------------------------------------------------------
   * Called by various functions if a clause is too big to parse or if
   * development becomes too expensive while there are less than 2
   * models.
   * 
   * Semantic guidance is now meaningless.  Behaviour defaults to
   * standard OTTER.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  record_event("\ndisable semantics after %d given clauses",
	       Stats[CL_GIVEN]);

  message(1,"WARNING: %s.\nDisabling semantic guidance "
	  "after %d given clauses\n", event, Stats[CL_GIVEN]);  

  auto_change_flag(stdout,SEMANTIC_GUIDANCE,0);
  auto_change_parm(stdout,PICK_GIVEN_RATIO,4);

  if (num_maxisets > 0) {// reset the SOS wts
    for (cl=Sos->first_cl; cl; cl=cl->next_cl) 
      cl->pick_weight = cl->length;
  }

  fflush(stdout);

}/*--disable_semantics()----------------------------------------------*/


void init_semantics(int argc, char *argv[])
  /*------------------------------------------------------------------
   * Called by OTTER fcn main() after input file read.
   *  
   * Initialises global variables and partition data structures;
   * Checks the command line to see if a findefile has been provided.
   * Calls init_finder() to prepare for the first model.  Finally,
   * builds an approriate semantic base.
   * -----------------------------------------------------------------*/
{
  int option, stat, maxiset;
  char filename[FILE_NAME_ARRAY]; 
  extern char *optarg;

  if (develop_semantics) return;

  *History = '\0'; 
  recording = true;

  // initialise principle global variables
  num_maxisets = 0;
  null_set_size = 0;
  develop_semantics = true;
  new_base_clauses.first_cp = NULL;
  new_base_clauses.last_cp = NULL;
  last_queue_theory = 0;

  // localise the following parameters
  last_set_index = tParms[MAX_NUM_MAXISETS].val+1; // one spare for replacement
  constraint_set = last_set_index + 1 ;

  // check to see if a finder file is provided
  strcpy(filename,"");
  while ((option = getopt (argc, argv, "f:")) != -1){
    switch( option ) { 
    case 'f':
      if (strlen(optarg) <= MAX_FILE_NAME)
	strcpy(filename,optarg);
      else abort_scott("filename longer than MAX_FILE_NAME characters.");
      break;
    case '?':
      fprintf(stderr,"\nusage: scott [-f finder_file] < stdin\n");
      fprintf(stderr,"%c",Bell);
      exit(1);
    }
  }

  // initialise partition data structures
  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
      givens_avoided_by[maxiset] = 0;
      last_num_updates[maxiset] = 0;
  }
  
  active[0] = true;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    active[maxiset] = false;  

  for (stat = 0; stat < MAX_TSTATS; stat++) tStats[stat] = 0;
  tStats[MAXISETS_AT_END_DEV] = -1;

  // Generate an initial model and set the base maxiset
  // This also checks the integrity of model constraints

  // First pass these parameters to the FINDER fragment
  set_min_prob_extn((float)fParms[MIN_PROB_EXTN].val);

  set_min_time_limit((int)(tParms[MIN_SEARCH_TIME].val/fParms[TOLERANCE].val));
  set_max_time_limit(tParms[MAX_SEARCH_TIME].val);
  set_tolerance(fParms[TOLERANCE].val);

  // maximum time limit for modelling finder file;
  set_time_limit(tParms[MAX_SEARCH_TIME].val);

  SCOTT_CLOCK_START(SEM_DEV_TIME);
  if (!init_finder(filename)){ 

    SCOTT_CLOCK_STOP(SEM_DEV_TIME);

    message(1,"WARNING: can't model the given semantics.\n"
	      "SCOTT is now running without semantic guidance");

    Flags[SEMANTIC_GUIDANCE].val = false;
    return;
  }

  set_time_limit(tParms[MIN_SEARCH_TIME].val);

  init_costs();
  
  // put here since init_clause_space() may be called twice
  semantic_base.first_cp = 
    semantic_base.last_cp = 
    new_base_clauses.first_cp = 
    new_base_clauses.last_cp = 
    conjecture.first_cp = 
    conjecture.last_cp = NULL; 

  SCOTT_CLOCK_STOP(SEM_DEV_TIME);

}/*--init_semantics()-------------------------------------------------*/


/*--------------------------------------------------------------------
 * Special macros for output_finder_stats().
 * -------------------------------------------------------------------*/

#define p_1d_stat(name, value)  fprintf(fp,"\n%-35s%5d",name,value)
#define p_2d_stat(name, value)\
{\
  fprintf(fp,"\n%-21s",name);\
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)\
    if (active[maxiset]) fprintf(fp,"%8d",value);\
}\

void output_finder_stats(FILE *fp, int stats_level)
  /*------------------------------------------------------------------
   * Called by OTTER fcn output_stats().  
   * 
   * Whenever OTTER outputs it's stats, appends appropriate FINDER
   * stats.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  fprintf(fp, "\n------------ semantic stats ------------\n");
  
  p_1d_stat("Semantic selections",                 tStats[SEMANTIC_SELECTIONS]);
  p_1d_stat("Final number of maxisets",            num_maxisets);

  if (tStats[MAXISETS_AT_END_DEV] > -1)
    p_1d_stat("Maxisets at end of development",      tStats[MAXISETS_AT_END_DEV]);

  p_1d_stat("Attempts at a new maxiset",           tStats[ATTEMPTS]);
  p_1d_stat("  to cover new null clauses",           tStats[COVER_ATTEMPTS]);
  p_1d_stat("  to refine choice",                    tStats[REFINE_ATTEMPTS]);

  p_1d_stat("New maxisets kept",                   tStats[KEPT]);
  p_1d_stat("  that covered a new null clause",      tStats[COVER_KEPTS]);
  p_1d_stat("  that refined choice",                 tStats[REFINE_KEPTS]);

  p_1d_stat("Maxisets forward subsumed",           tStats[FOR_SUBS]);
  p_1d_stat("Maxisets covering whole SOS",         tStats[SOS_COVERS]);
  p_1d_stat("Maxisets failing refinement",         tStats[REFINE_FAILS]);
  p_1d_stat("Existing maxisets back subsumed",     tStats[BACK_SUBS]);
  p_1d_stat("Existing maxisets culled",            tStats[CULLED]);
  p_1d_stat("Maxisets clashing with new base",     tStats[CLASHES]);

  fprintf(fp,"\n");
  p_1d_stat("Size of final semantic base",         tStats[SEM_BASE_SIZE]);
  p_1d_stat("Base clauses skipped",                tStats[BASE_CL_SKIPPED]);
  p_1d_stat("Base clauses added",                  tStats[BASE_CL_ADDED]);
  p_1d_stat("Base extension failures",             tStats[BASE_EXTENSION_FAILS]);

  fprintf(fp,"\n");
  p_1d_stat("Clauses given from input_first list", tStats[GIVENS_FROM_FIRST]);
  p_1d_stat("Clause placed in semantic queue",     tStats[CL_SEM_QUEUED]);
  p_1d_stat("Given clauses avoided",               givens_avoided);
  p_1d_stat("Number of pseudo null clauses",       null_set_size);
  p_1d_stat("Clauses marked false by time-limit",  get_time_outs());
  p_1d_stat("  while still in test",               get_time_outs_in_test());
  p_1d_stat("Time limit at end of search",         get_time_limit());
  p_1d_stat("Clauses modelled by retesting",       tStats[SUCCESSFUL_RETESTS]);

  fprintf(fp,"\n");
  p_1d_stat("Fails in persist()",                  tStats[FAILS_IN_PERSIST]);
  p_1d_stat("   time saved strictly increasing",  tStats[PFAIL_TIME_INC]);
  p_1d_stat("   time saved strictly decreasing",  tStats[PFAIL_TIME_DEC]);
  p_1d_stat("   cost vs time left",               tStats[PFAIL_COST_VS_TIME]);
  p_1d_stat("   cost vs time accrued",            tStats[PFAIL_TIME_ACCRUED]);
  p_1d_stat("   probability strictly decreasing", tStats[PFAIL_PROB_DEC]);
  p_1d_stat("   cost vs probability",             tStats[PFAIL_COST_VS_PROB]);
  p_1d_stat("   cost vs expected return",         tStats[PFAIL_COST_VS_RETURN]);
  fprintf(fp,"\n");

  if (Flags[SEMANTIC_GUIDANCE].val && num_maxisets > 0) {
    p_2d_stat("Maxiset index",     maxiset);    
    fprintf(fp,"\n---------------------");
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) 
      if (active[maxiset]) fprintf(fp,"--------");
    p_2d_stat("Last maxiset size",    get_clauses_modelled(maxiset));
    p_2d_stat("Givens avoided by",    givens_avoided_by[maxiset]);
    p_2d_stat("Models searched for",  get_searches(maxiset));
    p_2d_stat("Models found",         get_successful_searches(maxiset));
    p_2d_stat("Max backtracks needed",get_max_backtracks_in(maxiset));
    p_2d_stat("Max time needed",      get_max_time_in(maxiset));
    p_2d_stat("Final domain size",    get_last_default_size(maxiset));
    fprintf(fp,"\n");
    
    print_partition("at end of search");  

    debug_partition("at end of search");
    debug_scan("at end of search");

    message(2,"");
  }
  message(2,"");
  output_weight_distn();
  output_variable_distn();
  message(2,"");
  output_history();
  
}/*--output_finder_stats()--------------------------------------------*/


long scott_clock_val(int c)
  /*------------------------------------------------------------------
   * Called by cost_per_attempt() and cost_per_given().
   * 
   * Returns accumulated time (in milliseconds) of the given clock.
   * Clock need not be stopped.
   * -----------------------------------------------------------------*/
{
    long sec, usec, i, j;

    i = (Clocks[c].accum_sec * 1000) + (Clocks[c].accum_usec / 1000);
    if (Clocks[c].curr_sec == -1)
	return(i);
    else {
	SCOTT_CPU_TIME(sec, usec)
	j = ((sec - Clocks[c].curr_sec) * 1000) +
	    ((usec - Clocks[c].curr_usec) / 1000);
	return(i+j);
	}
}/*--scott_clock_val()------------------------------------------------*/

void message(int priority, char *msg, ...)
  /*------------------------------------------------------------------
   * Called by various functions...
   * 
   * Prints the given message to the screen iff priority is greater
   * than or equal to the VERBOSITY parameter.  -- msg may be a std
   * format string, and in this case may be followed by arbitrary
   * arguments.
   * 
   * If the priority is less than 2 then the terminal bell is rung.
   * -----------------------------------------------------------------*/
{
  static boolean newline = false;

  if (priority <= tParms[VERBOSITY].val) {
    va_list args;
    
    if (strcmp(msg, "") == 0) {
      newline = true;
      return;
    }
    
    if (newline) printf("\n");

    va_start(args, msg);
    vprintf(msg, args);
    printf("\n");
    fflush(stdout);
    
    if (priority < 2) {

      if (newline) fprintf(stderr,"\n");

      fprintf(stderr,"%c",Bell);
      vfprintf(stderr, msg, args);
      fprintf(stderr,"\n");
    }
    newline = false;

    fflush(stdout);
    va_end(args);
  }
} /*--message()-------------------------------------------------------*/

void abort_scott(char *msg, ...)
    /*------------------------------------------------------------------
   * Called by various functions...
   * 
   * An error has been detected.  the terminal bell is rung, the given
   * message is printed to the screen, and the program exits -- msg
   * may be a std format string, and in this case may be followed by
   * arbitrary arguments.
   * -----------------------------------------------------------------*/
{
  va_list args;

  va_start(args, msg);

  fprintf(stderr,"\nERROR: ");
  fprintf(stderr,"%c", Bell);
  vfprintf(stderr, msg, args);
  fprintf(stderr, "\n");

  fprintf(stdout,"\nERROR: ");
  vfprintf(stdout, msg, args);
  fprintf(stdout, "\n");

  va_end(args);

  fprintf(stdout,"\nEvents preceding the error:\n\n");
  output_history();

  exit(1);

}/*--abort_scott()----------------------------------------------------*/

void record_event(char *event, ...)
{

  if (recording) {
    va_list args;
    char tmp[10];

    *tmp = '\0';
    
    va_start(args, event);
    vsprintf(tmp,event,args);
    strncat(History,tmp,MAX_HISTORY);
    va_end(args);
  }
}/*--record_event()---------------------------------------------------*/

void output_history()
{
  message(2,"----> Start Development history:\n%s\n\n"
	 "4<---- End development history\n",History);

}/*--output_history()-------------------------------------------------*/

/*--------------------------------------------------------------------
 * The following group of functions simply return the value of a
 * variable that would otherwise be out of scope.
 * -------------------------------------------------------------------*/
// Called Foutput.c:fprint_model()
int get_num_maxisets() { return num_maxisets; }

// Called by clause.c:print_clause() and process.c:pre_process()
boolean in_constraint_set(struct clause *cl) {  return cl->in_set[constraint_set];}

/*--------------------------------------------------------------------*/
