/*------------------------------------------------------------------------*
 *                         Sauto.c                                        *
 *                                                                        *
 *                 SCOTT's autonomous mode                                *
 *                                                                        *
 *------------------------------------------------------------------------*/

#include "SCOTT.h"

#define MAX_PROPERTIES 20

#define NUM_CL         0
#define NUM_POS_CL     1
#define NUM_NEG_CL     2
#define NUM_MIX_CL     3
#define NUM_EQ_CL      4
#define NUM_UNIT_CL    5
#define NUM_GRD_CL     6
#define NUM_HORN_CL    7
#define MAX_LITS       8
#define MIN_LITS       9
#define MAX_WT        10
#define MIN_WT        11
#define MAX_DIST_VARS 12
#define MIN_DIST_VARS 13
    
static int property[MAX_PROPERTIES];

boolean all_clauses(int (*property)()) 
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Returns TRUE iff every clause
   * on the USABLE, SOS, and INPUT_FIRTS lists have the given
   * property.
   * -----------------------------------------------------------------*/
{
    struct clause *c;
    boolean test = true;
    
    for (c=Usable->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    for (c=Sos->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    for (c=First->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    return test;

} /*--all_clauses()---------------------------------------------------*/

boolean at_least_one_clause(int (*property)())
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Returns TRUE iff at least one
   * clause on the USABLE, SOS, and INPUT_FIRTS lists has the given
   * property.
   * -----------------------------------------------------------------*/
{
    struct clause *c;
    boolean test = false;
    
    for (c=Usable->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    for (c=Sos->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    for (c=First->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    return test;

} /*--at_least_one_clause()-------------------------------------------*/

int get_max(int (*property)())
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Returns maximum value across
   * the USABLE, SOS, and INPUT_FIRST lists of the given property.
   * -----------------------------------------------------------------*/
{    
  struct clause *c;
  int next, max = 0;

  for (c=Usable->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  for (c=Sos->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  for (c=First->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  return max;

}/*get_max()----------------------------------------------------------*/

int get_min(int (*property)())
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Returns minimum value across
   * the USABLE, SOS, and INPUT_FIRST lists of the given property.
   * -----------------------------------------------------------------*/
{    
  struct clause *c;
  int next, min = MAX_INT;

  for (c=Usable->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  for (c=Sos->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  for (c=First->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  return min;

}/*get_min()----------------------------------------------------------*/

int get_num(int (*property)())
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Returns number of clauses on
   * the USABLE, SOS, and INPUT_FIRST lists of the given property.
   * -----------------------------------------------------------------*/
{    
  struct clause *c;
  int num = 0;

  for (c=Usable->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  for (c=Sos->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  for (c=First->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  return num;

}/*get_num()----------------------------------------------------------*/

int mix_clause(struct clause *c)
{
  return (!(pos_clause(c) || neg_clause(c)));

}/*mixed_clause()-----------------------------------------------------*/


int clause_weight(struct clause *c)
{
  return (weight_cl(c, Weight_pick_given_index));

}/*clause_weight()----------------------------------------------------*/

int pos_ground_clause(struct clause *c)
{
    return (pos_clause(c) && ground_clause(c));

}/*pos_ground_clause()---------------------------------------------*/

int neg_ground_clause(struct clause *c)
{
    return (neg_clause(c) && ground_clause(c));

}/*pos_ground_clause()---------------------------------------------*/

int equality_axiom(struct clause *c)
  /*------------------------------------------------------------------
   * Called by auto_set_conjecture() Returns true iff "c" is a
   * symmetry or reflexivity clasue for equality.
   * ------------------------------------------------------------------*/
{
  if (symmetry_clause(c)) return 1;
  if (num_literals(c) != 1) return 0;
  if (!eq_lit(c->first_lit)) return 0;
  else {
    struct term *atom;
    
    atom = c->first_lit->atom;
    
    if (!atom->farg->argval->type == VARIABLE)
      return 0;
    
    if (!atom->farg->narg->argval->type == VARIABLE)
      return 0;
    
    if (!term_ident(atom->farg->argval,
		    atom->farg->narg->argval))
      return 0;
  }
  return 1;

}/*equality_axiom() --------------------------------------------------*/


void automatic_3_settings()
  /*------------------------------------------------------------------
   * Called by read_all_input().
   * 
   * SCOTTs very own autonomous mode.
   * -----------------------------------------------------------------*/
{
  boolean propositional, horn, equality, symmetry;

  if (Passive->first_cl)
    abend("Passive list not accepted in auto3 mode.");
  else if (Demodulators->first_cl)
    abend("Demodulators list not accepted in auto3 mode.");
  else if (Weight_pick_given || Weight_purge_gen || Weight_terms)
    abend("Weight lists not accepted in auto3 mode.");
  
  printf("\n------------> automatic settings\n");
  
  /* Find out some basic properties. */
  
  propositional = all_clauses(propositional_clause);
  horn = all_clauses(horn_clause);
  equality = at_least_one_clause(equality_clause);
  symmetry = at_least_one_clause(symmetry_clause);
  
  printf("\nInput clauses form a %s%s set%s%s.\n\n",
	 (propositional ? "propositional," : ""), 
	 (horn ? "horn" : "mixed"), 
	 (equality ? " with equality" : " without equality"), 
	 (symmetry ? " and a symmetry clause" : "")
	 );
  
  set_properties();
  
  if (sos_has_pos_nonground())
    printf("\nSos has positive nonground clause; therefore it is not changed.\n");
  else {
    printf("\nEvery positive clause in sos is ground (or sos is empty);\n");
    printf("therefore we move all positive usable clauses to sos.\n");
    move_clauses(pos_clause, Usable, Sos);
  }
  
  if (propositional) {
    printf("\nAll clauses are propositional; therefore we set the\n");
    printf("propositional flag and use ordered hyperresolution.\n");
    auto_change_flag(stdout, PROPOSITIONAL, 1);
    auto_change_flag(stdout, HYPER_RES, 1);
    return;
  }
      
  /* nonpropositional */
  
  if (property[MAX_LITS] == 1) {
    printf("\nSetting pick_given_ratio to 2, ");
    printf("because all clauses are units.\n");
    auto_change_parm(stdout, PICK_GIVEN_RATIO, 4);
  }
  else {
    /* Nonunit */
    printf("\nSetting hyper_res, because there are nonunits.\n");
    auto_change_flag(stdout, HYPER_RES, 1);
    
    if (equality || !horn) {
      printf("\nSetting ur_res, because this is a nonunit set containing\n");
      printf("either equality literals or non-Horn clauses.\n");
      auto_change_flag(stdout, UR_RES, 1);
    }
    
    if (horn) {
      printf("\nClearing order_hyper, because all clauses are Horn.\n");
      auto_change_flag(stdout, ORDER_HYPER, 0);
    }
  
    if (!horn) {
      /* Non-Horn */
      printf("\nSetting factor and  unit_deletion, ");
      printf("because there are non-Horn clauses.\n");
      auto_change_flag(stdout, FACTOR, 1);
      auto_change_flag(stdout, UNIT_DELETION, 1);
    }
  }

  if (equality) {
    printf("\nEquality is present, so we set the knuth_bendix flag.\n");
    auto_change_flag(stdout, KNUTH_BENDIX, 1);
    if (property[MAX_LITS] > 1) {
      printf("\nAs an incomplete heuristic, we paramodulate with units only.\n");
      auto_change_flag(stdout, PARA_FROM_UNITS_ONLY, 1);
      auto_change_flag(stdout, PARA_INTO_UNITS_ONLY, 1);
    }
    if (symmetry) {
      printf("\nThere is a clause for symmetry of equality, so it is\n");
      printf("assumed that equality is fully axiomatized; therefore,\n");
      printf("paramodulation is disabled.\n\n");
      auto_change_flag(stdout, PARA_FROM, 0);
      auto_change_flag(stdout, PARA_INTO, 0);
    }
  }
  
  
}/*--automatic_3_settings()-------------------------------------------*/


void auto_set_sos()
  /*------------------------------------------------------------------
   * Called by read_all_input() *after* auto_set_conjecture().  
   * 
   * Moves SOS clauses to the USABLE iff we can show that the
   * resulting list is consistent and does not interfere with the
   * completeness of any inference rules.
   * ------------------------------------------------------------------*/
{
  struct clause *cl, *next, *first_sos;
  struct clause_dptr *cp, *next_cp;
  char finder_cl[CLAUSE_STRING_ARRAY];
  int new_time_limit = 0;
  boolean first;
  
  // sos strategy conflicts with pure paramodulation
  if (property[NUM_UNIT_CL] == property[NUM_EQ_CL]
      && Flags[KNUTH_BENDIX].val) return;

  message(2,"");
  message(2,"------------> checking set of support strategy");
  message(2,"");

  // Use maxiset position 1 for this testing
  // and set maximum time limit for modelling input clauses
  set_time_limit(tParms[MAX_SEARCH_TIME].val);

  // model each conjecture clause
  cp = conjecture.first_cp; 
  while (cp){
    next_cp = cp->next;
    cl = cp->c;
    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (model(finder_cl, 1, false)){
      cl->in_base = neither;
    }
    else {
      delete_cl_ext(cl,&conjecture);
      cl->in_base = false;
    }
    cp = next_cp;
  }

  // next model the USABLE list

  first = true;
  first_sos = Sos->first_cl;
  cl=Usable->first_cl; 
  while (cl) { 
    next = cl->next_cl;
    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (!model(finder_cl, 1, false)){
      if (first) message(2,"Moving the following seemingly inconsistent SOS"
			 " clauses to the USABLE list..");
      print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      prepend_cl(Sos,cl);
    }
    cl = next;
  }

  message(2,"");

  first = true;
  cl=Sos->first_cl; 
  while (cl) { 
    next = cl->next_cl;
    
    // dont move units if using hyper_res
    if (unit_clause(cl) 
	&& (Flags[HYPER_RES].val || Flags[NEG_HYPER_RES].val)
	){
      cl = next;
      continue;
    }
    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (model(finder_cl, 1, false)) {
      if (first) message(2,"Moving the following consistent SOS"
			 " clauses to the USABLE list..");
      print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      append_cl(Usable,cl);
    }
    cl = next;
  }
   
  // this should give us an appropriate starting time_limit
  new_time_limit = get_max_time_in(1);
  if (new_time_limit < tParms[MIN_SEARCH_TIME].val)
    set_time_limit(tParms[MIN_SEARCH_TIME].val);
  else 
    set_time_limit(new_time_limit);
  
  // reinit tempory maxiset position 1
  init_sizes(1);
  init_theory(1);

  message(2,"");
  
}/*auto_set_sos()-----------------------------------------------------*/

void auto_set_conjecture()
  /*------------------------------------------------------------------
   * Called by read_all_input() *before* auto_set_sos().
   * 
   * Determines the most likely properties of a conjecture clause and
   * ensures that the input first list is set appropriately. Set the
   * extenral list "conjecture".
   * -----------------------------------------------------------------*/
{
  int (*conj_clause)();
  struct clause *cl, *next;
  boolean first;

  conj_clause = NULL;
  
  message(2,"");
  message(2,"------------> checking conjecture clause");
  message(2,"");

  // set the explicit conjecture clauses
  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    if (cl->in_base != neither ){
      cl->in_base = neither;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[SEM_BASE_SIZE]++;
    }
  }
  /*
  // hope this works
  for (cl=Demodulators->first_cl; cl; cl = cl->next_cl) {
    if (cl->in_base != neither ){
      cl->in_base = neither;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[SEM_BASE_SIZE]++;
    }
  }
  */
  if (property[NUM_POS_CL] > property[NUM_NEG_CL]){
    if (get_num(neg_ground_clause)){
      message(2,"Conjecture is most likely a negative ground clause");
      conj_clause = neg_ground_clause;
    }
    else {
      message(2,"Conjecture is most likely a negative clause");
      conj_clause = neg_clause;
    }
  }
  else if (property[NUM_POS_CL] < property[NUM_NEG_CL]){
    if (get_num(pos_ground_clause)){
      message(2,"Conjecture is most likely a positive ground clause");
      conj_clause = pos_ground_clause;
    }    
    else {
      message(2,"Conjecture is most likely a positive clause");
      conj_clause = pos_clause;
    }
  }

  if (!conj_clause) return;



  message(2,"");
  first = true;
  cl=Sos->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if ((*conj_clause)(cl)) {
      if (first) 
	message(2,"Moving the following conjecture "
		"clauses to the INPUT_FIRST list...");
      print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      append_cl(First,cl);
      cl->in_base = neither;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[SEM_BASE_SIZE]++;
    }
    cl = next;
  }
  
  message(2,"");
  first = true;
  cl=First->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if (!(*conj_clause)(cl)) {
      if (first) 
	message(2,"Moving the following non conjecture "
		"clauses to the SOS list...");
      print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      append_cl(Sos,cl);
      cl->in_base = false;
      delete_cl_ext(cl,&conjecture); 
      tStats[SEM_BASE_SIZE]++;
    }
    cl = next;
  }

  message(2,"");
  first = true;
  cl=Usable->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if ((*conj_clause)(cl)) {
      if (first) 
	message(2,"Marking the following USABLE clause "
		"as conjectures...");
      print_clause(stdout,cl);
      first = false;
      cl->in_base = neither;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[SEM_BASE_SIZE]++;
    }
    cl = next;
  }
  if (!conjecture.first_cp 
      || !conjecture.first_cp->c) 
    message(1,"WARNING: no conjecture clause has been established.\n");
  
}/*auto_set_conjecture()--------------------------------------------- */


void set_properties()
  /*------------------------------------------------------------------
   * Called by automatic_3_settings().  Set clause space properties.
   * -----------------------------------------------------------------*/
{
  property[MAX_LITS] = get_max(num_literals);
  property[MIN_LITS] = get_min(num_literals);
  property[MAX_WT] = get_max(clause_weight);
  property[MIN_WT] = get_min(clause_weight);
  property[MAX_DIST_VARS] = get_max(distinct_vars);
  property[MIN_DIST_VARS] = get_min(distinct_vars);
  
  property[NUM_CL] = Stats[SOS_SIZE] +Stats[USABLE_SIZE];
  property[NUM_POS_CL] = get_num(pos_clause);
  property[NUM_NEG_CL] = get_num(neg_clause);
  property[NUM_MIX_CL] = get_num(mix_clause);
  property[NUM_EQ_CL] = get_num(equality_clause);
  property[NUM_HORN_CL] = get_num(horn_clause);
  property[NUM_UNIT_CL] = get_num(unit_clause);
  property[NUM_GRD_CL] = get_num(ground_clause);
  
  printf("There are %2d clauses of which ...\n"
	 "  %2d are positive,\n"
	 "  %2d are negative,\n"
	 "  %2d are mixed,\n"  
	 "  %2d are horn,\n"
	 "  %2d are unit,\n"
	 "  %2d are ground,\n"	   
	 "  %2d are equality.\n\n"
	 "Clause weight ranges from %d to %d.\n"  
	 "Variable count ranges from %d to %d.\n" 
	 "Literal count ranges from %d to %d.\n",
	 property[NUM_CL], 
	 property[NUM_POS_CL], 
	 property[NUM_NEG_CL], 
	 property[NUM_MIX_CL], 
	 property[NUM_HORN_CL], 
	 property[NUM_UNIT_CL], 
	 property[NUM_GRD_CL], 
	 property[NUM_EQ_CL],
	 property[MIN_WT], property[MAX_WT],
	 property[MIN_DIST_VARS], property[MAX_DIST_VARS],
	 property[MIN_LITS], property[MAX_LITS]);
  
}/*set_properties()---------------------------------------------------*/

void check_eq_settings()
{
  if (at_least_one_clause(equality_clause) 
      && property[MAX_LITS] > 1){
    printf("\nAs an incomplete heuristic, we paramodulate with units only.\n");
    auto_change_flag(stdout, PARA_FROM_UNITS_ONLY, 1);
    auto_change_flag(stdout, PARA_INTO_UNITS_ONLY, 1);
  }
}
