/*------------------------------------------------------------------------*
 *                        Sclause_space.c                                 *
 *                                                                        *
 *       SCOTT functions that operate on the clause space                 *
 *                                                                        *
 *------------------------------------------------------------------------*/

#include "SCOTT.h"

#if  (MAX_VARS < (VARMAX-2))
#define SCOTT_MAX_VARS MAX_VARS   // OTTER maximum
#else
#define SCOTT_MAX_VARS (VARMAX-2) // FINDER maximum
#endif

static struct clause_dptr *clause_dptr_avail;
static long clause_dptr_gets, clause_dptr_frees, clause_dptr_avails;

void init_clause_dptr()  
  /*------------------------------------------------------------------
   * Called by init_clause_space().  Initialise dynamic memory.
   * -----------------------------------------------------------------*/
{
  clause_dptr_frees = 0;
  clause_dptr_avails = 0;
  clause_dptr_gets = 0;
  clause_dptr_avail = NULL;

}/*--init_clause_space()----------------------------------------------*/

struct clause_dptr *get_clause_dptr()
  /*------------------------------------------------------------------
   * Called by prepend_cl_ext() and append_cl_ext().  Allocates
   * appropriate memory block and returns a pointer to the block.
   * -----------------------------------------------------------------*/
{
    struct clause_dptr *p;

    clause_dptr_gets++;
    if (clause_dptr_avail == NULL)
	p = (struct clause_dptr *) tp_alloc((int) sizeof(struct clause_dptr));
    else {
	clause_dptr_avails--;
	p = clause_dptr_avail;
	clause_dptr_avail = clause_dptr_avail->next;
	}
    p->next = NULL;
    p->prev = NULL;
    p->c = NULL;
    return(p);

}/*--get_clause_dptr()------------------------------------------------*/


void free_clause_dptr(struct clause_dptr *p)
  /*------------------------------------------------------------------
   * Called by delete_cl_ext().  Frees the memory associated with the
   * given pointer.
   * -----------------------------------------------------------------*/
{
    clause_dptr_frees++;
    clause_dptr_avails++;
    p->next = clause_dptr_avail;
    clause_dptr_avail = p;

}/*--free_clause_dptr()-----------------------------------------------*/


void delete_cl_ext(struct clause *c, struct external_list *elist)
  /*------------------------------------------------------------------
   * Called by rem_from_list().  Deletes the pointer to c in the given
   * doubly linked list
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *cp, *prev_cp;

  cp = elist->first_cp;
  prev_cp = NULL; 

  while (cp && cp->c != c){
    prev_cp = cp;   // remember the predecessor
    cp = cp->next;
  }
  
  if (!cp || cp->c != c) 
    abort_scott("delete_cl_ext() could not find clause "
		"in the given list of pointers.");

  if (!prev_cp) // first in the list
    elist->first_cp = cp->next;
  else 
    prev_cp->next = cp->next; // skip forward over cp

  if (!cp->next)  // last in the list
    elist->last_cp = prev_cp;
  else 
    cp->next->prev = prev_cp; // skip backward over cp

  free_clause_dptr(cp);
  
}/*--delete_cl_ext()--------------------------------------------------*/

void prepend_cl_ext(struct clause *c, struct external_list *elist)
  /*------------------------------------------------------------------
   * Called by start_scan(), scan_sos(), scan_base(),
   * sos_scan_insert()
   * 
   * Establishes a new pointer to c at or near the begining of the
   * given pointer list.  If the clause has only one model an attempt
   * is made to ensure that models alternate.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *new_cp;
  
  new_cp = get_clause_dptr();
  new_cp->c = c;

  // list is empty
  if(!elist->last_cp) {
    elist->first_cp = elist->last_cp = new_cp;
    return;
  }
  
  if (c->num_sets == 1) { // alternate maxisets

    struct clause_dptr *prev_cp, *next_cp;

    prev_cp = NULL;
    next_cp = elist->first_cp;
    while (next_cp) {
      // ensure next clause has different models
      if (strncmp(next_cp->c->in_set,c->in_set,MAXISET_ARRAY) != 0){
	if (prev_cp) prev_cp->next = new_cp;
	else elist->first_cp = new_cp;
	next_cp->prev = new_cp;
	new_cp->next = next_cp;
	new_cp->prev = prev_cp;
	return;
      }
      prev_cp = next_cp;
      next_cp = next_cp->next;
    }
  }
  // otherwise, or if this fails, stick it in the front
  new_cp->next = elist->first_cp;
  elist->first_cp->prev = new_cp;
  elist->first_cp = new_cp;

}/*--prepend_cl_ext()-------------------------------------------------*/


void append_cl_ext(struct clause *c, struct external_list *elist)
  /*------------------------------------------------------------------
   * Called by start_scan(), scan_sos(), scan_base(),
   * sos_scan_insert(), init_clause_space().
   * 
   * Establishes a new pointer to c at or near the end ofthe given
   * pointer list.  If the clause has only one model, an attempt is
   * made to ensure that models alternate.
   * -----------------------------------------------------------------*/
{
  struct clause_dptr *new_cp;
  
  new_cp = get_clause_dptr();
  new_cp->c = c;

  // list is empty
  if(!elist->last_cp) {
    elist->first_cp = elist->last_cp = new_cp;
    return;
  }

  if (c->num_sets == 1) { // alternate maxisets

    struct clause_dptr *prev_cp, *next_cp;

    //scan backward
    next_cp = NULL;
    prev_cp = elist->last_cp;
    while (prev_cp) {
      // ensure prev clause has different models
      if (strncmp(prev_cp->c->in_set,c->in_set,MAXISET_ARRAY) != 0){
	if (next_cp) next_cp->prev = new_cp;
	else elist->last_cp = new_cp;
	prev_cp->next = new_cp;
	new_cp->prev = prev_cp;
	new_cp->next = next_cp;
	return;
      }
      next_cp = prev_cp;
      prev_cp = prev_cp->prev;
    }
  }

  // otherwise or if this fails, stick it in the back
  new_cp->prev = elist->last_cp;
  elist->last_cp->next = new_cp;
  elist->last_cp = new_cp;
  
}/*--append_cl_ext()--------------------------------------------------*/



void sos_scan_insert(struct clause *c)
  /*------------------------------------------------------------------
   * Called by scott_insert() adn init_clause_space().
   * 
   * Insert the given clause into the appropriate external lists.  New
   * clauses are placed at the start of the list so that the scan
   * order is always changing.
   * -----------------------------------------------------------------*/
{
  int maxiset;
  
  if (!develop_semantics) return; // this structure is no longer needed
  
  prepend_cl_ext(c,&sos_scan[c->num_sets]);
  
  // insert in failed_clauses[] too!
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset] && !c->in_set[maxiset]) 
      prepend_cl_ext(c,&failed_clauses[maxiset]);
  
}/*--sos_scan_insert()------------------------------------------------*/


void sos_scan_delete(struct clause *c)
  /*------------------------------------------------------------------
   * Called by scott_delete().
   * 
   * Removes the given clause from any external lists that contain it.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  if (!develop_semantics) return; // this structure is no longer needed
  
  delete_cl_ext(c,&sos_scan[c->num_sets]);

  // remove from failed_clauses[] too!
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset] && !c->in_set[maxiset]) 
      delete_cl_ext(c,&failed_clauses[maxiset]);

}/*--sos_scan_delete()------------------------------------------------*/


void declare_soft_equality()  
  /*------------------------------------------------------------------
   * Called by init_clause_space().
   *
   * Something experimental.  adjust the base maxiset to force either
   * anti-symmetric or non-symmetric.  This assumes that we are using
   * a special equality definition that does not contain the symmetry
   * axiom for equality.
   * -----------------------------------------------------------------*/
{
  //---- anti-symmetric equality
  // model("equal(A,B), equal(B,A) -> $false.", 0, false);
  
  //---- non-symmetric equality
  add_declaration("constant scott_a default.");
  add_declaration("constant scott_b default.");
  model("equal(scott_a,scott_b), equal(scott_b,scott_a) -> $false.", 0, false);
  
}/*--declare_soft_equality()------------------------------------------*/


void declare_hard_equality()  
  /*------------------------------------------------------------------
   * Called by init_clause_space().  Insist that all models of
   * equality are the identity reklation
   * -----------------------------------------------------------------*/
{
  init_theory(0); // hack to get rid of any existing equality definition
  model("equal(A,B) = (A = B).",0,false);

}/*--declare_hard_equality()------------------------------------------*/


void set_base() 
  /*------------------------------------------------------------------
   * Called by init_semantics().
   * 
   * Sets an initial semantic base consisting of all th usable list
   * clauses.  This must come after the initial Usable list has been
   * set by auto_set_sos().
   * -----------------------------------------------------------------*/
{
  struct clause *cl;

  // set initial base theory
  for (cl = Usable->first_cl; cl; cl=cl->next_cl) {
    if (cl->in_base != true) {
      cl->in_base = true;
      append_cl_ext(cl,&semantic_base); 
      tStats[SEM_BASE_SIZE]++;
    } 
  }

  /*
  // add new dynamic demodulators to the conjecture
  for (cl = Demodulators->first_cl; cl; cl=cl->next_cl) {
    scott_insert_conjecture(cl);
  }
  */

}/*set_base()---------------------------------------------------------*/


void init_clause_space()
  /*------------------------------------------------------------------
   * Called by init_semantics().
   * 
   * Initialises external clause lists sos_scan[], failed_clauses[],
   * semantic_base, and conjecture.  Since FINDER was not active
   * during input processing, also ensures clause lengths have been
   * recorded.  This must come before auto_set_conjecture() and
   * auto_set_sos(), since these require clauses to have been
   * integrated ito there respective lists.
   * -----------------------------------------------------------------*/
{
  int maxiset;
  struct clause *cl;

  init_clause_dptr();

  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
    sos_scan[maxiset].first_cp = 
      sos_scan[maxiset].last_cp = 
      failed_clauses[maxiset].first_cp = 
      failed_clauses[maxiset].last_cp = NULL;
  }

  // set weights of initial clause space
  for (cl = Usable->first_cl; cl; cl=cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }
  
  for (cl=Sos->first_cl; cl; cl=cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
    sos_scan_insert(cl); // initial scan
  }

  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }

}/*--init_clause_space() ---------------------------------------------*/


boolean make_finder_clause(char *finder_clause, struct clause *otter_clause)
  /*------------------------------------------------------------------
   * Called by model_list() and add_semantic_wt().  
   * 
   * Given an OTTER clause, creates a corresponding FINDER clause.
   * Return TRUE iff the clause is successfully converted.
   * -----------------------------------------------------------------*/
{
  struct literal *lit;
  int separator, num_vars;

  *finder_clause = '\0'; /* start with an empty FINDER clause */
  
  num_vars = distinct_vars(otter_clause);
  if (num_vars >= SCOTT_MAX_VARS) {
    // assume the clause is shorter than the current limits
    message(0,"WARNING: clause has too many (%d) variables", 
	    num_vars);
    message(0,"Resetting distinct variable limit to %d\n",
	    num_vars-1);    
    Parms[MAX_DISTINCT_VARS].val = num_vars-1;
    cull_clauses_over(num_vars-1,distinct_vars);
    return false;
  }

  /* convert negative literals */
  for (lit=otter_clause->first_lit,separator=false;lit;lit=lit->next_lit)
    if (!lit->sign) {
      if (separator)
	strcat(finder_clause, ", ");
      if (!add_finder_term(finder_clause, lit->atom)) {
	// assume the clause is shorter than the current limits
	message(0,"WARNING: clause is too long (%d characters)",
		CLAUSE_STRING_ARRAY);
	message(0,"Resetting weight limit to %d\n",otter_clause-1);    
	Parms[MAX_WEIGHT].val = otter_clause->pick_weight-1;
	cull_clauses_over(otter_clause->pick_weight-1,clause_pick_weight);
	return false;
      }
      else separator = true;      
    }
  
  strcat(finder_clause, " -> ");
  
  /* convert positive literals */
  for (lit=otter_clause->first_lit,separator=false;lit;lit=lit->next_lit)
    if (lit->sign) {
      if (separator)
	strcat(finder_clause, ", ");
      if (!add_finder_term(finder_clause, lit->atom)) {
	// assume the clause is shorter than the current limits
	message(0,"WARNING: clause is too long (%d characters)",
		CLAUSE_STRING_ARRAY);
	message(0,"Resetting weight limit to %d\n",otter_clause-1);    
	Parms[MAX_WEIGHT].val = otter_clause->pick_weight-1;
	cull_clauses_over(otter_clause->pick_weight-1,clause_pick_weight);
	return false;
      }
      else separator = true;
    }
  strcat(finder_clause, ".");

  return true;
}/*--make_finder_clause()---------------------------------------------*/


boolean add_finder_term(char *finder_clause, struct term *otter_term)
  /*------------------------------------------------------------------
   * Called by make_finder_clause().  
   * 
   * Converts an OTTER term into a FINDER term and appends to a
   * partial finder_clause.  Recursively checks that the clause hasn't
   * grown too large.  Return TRUE iff the clause is successfully
   * converted.
   * -----------------------------------------------------------------*/
{
  struct rel *r;
  char variable_index[VARMAX];
  static boolean clause_too_big = false;

  if (clause_too_big) return false;

  /* exit if s is too big to append the next name to */
  if (strlen(finder_clause) >= (CLAUSE_STRING_ARRAY - FUNCTION_NAME_ARRAY)){
    clause_too_big =true;
    return false;
  }
  if (otter_term->type == NAME){
    strcat(finder_clause, sn_to_str(otter_term->sym_num));
  }
  else if (otter_term->type == VARIABLE) {
    strcat(finder_clause, "V");
    // now we need to turn the number into a string
    sprintf(variable_index, "%d", otter_term->varnum);
    strcat(finder_clause, variable_index);
  }
  else {
    strcat(finder_clause, sn_to_str(otter_term->sym_num));
    strcat(finder_clause, "(");
    for (r = otter_term->farg; r; r = r->narg) {
      if(!add_finder_term(finder_clause, r->argval)) return false ;
      if (r->narg)
	strcat(finder_clause, ",");
    }
    strcat(finder_clause, ")");
  }
  return true;

}/*-add_finder_term()-------------------------------------------------*/


void in_set_to_str(char *in_set, struct clause *cl)
  /*------------------------------------------------------------------ 
   * Called by extend_base() and print_in_set().
   *
   * Copies the "in_set" field of "cl" across to the string pointer to
   * by "in_set", transferring appropriate character codes.
   * -----------------------------------------------------------------*/
{
    int maxiset, count;

    *in_set = '(';
    in_set++;
    for (maxiset = 0, count= 0; 
	 maxiset <= last_set_index && count <= num_maxisets; 
	 maxiset++, in_set++) {
      if (!active[maxiset]) {
	*in_set = 'x';
	continue;
      }
      else if (cl->in_set[maxiset]) {
	if (cl->in_base == true) *in_set = 'b';
	else if (cl->in_base == neither) *in_set = 'n';
	else *in_set = '1';
      }
      else *in_set = '0';
      count++;
    }
    *in_set = ')';
    
    if (Flags[SEM_RES].val) {
      in_set++; *in_set = ' '; in_set++;
      if (cl->in_set[constraint_set]) *in_set = 'T';
      else *in_set = 'F';
    }
    
    *(in_set+1)  = '\0';

}/*in_set_to_str()----------------------------------------------------*/


boolean in_set_identical(struct clause *c1, struct clause *c2)
  /*------------------------------------------------------------------
   * Called by shuffle_ext_list().  Returns true iff in_set fileds of
   * the two clauses are identical.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
    if (c1->in_set[maxiset] != c2->in_set[maxiset]) 
      return false;
  }
  return true;

}/*in_set_identical()-------------------------------------------------*/


/*================== DEBUGGING TOOLS =================================*/

void print_sos_scan()
{
  int list;
  
  for (list = 0; list <= num_maxisets; list++) {
    printf("\n%d | ",list);
    print_ext_list(&sos_scan[list]);
  }
  printf("\n\n");
  fflush(stdout);
  
} /* print_sos_scan() ------------------------------------------------*/


void print_failed_clause_nos()
{
  int num_sets;
  
  for (num_sets = 0; num_sets<=num_maxisets; num_sets++){
    printf("\n%d | ",num_sets);
    print_ext_list(&failed_clauses[num_sets]);
  }
  printf("\n\n");
  fflush(stdout);
  
} /* print_failed_clauses() ------------------------------------------*/


void print_conjecture_nos()
{    
  struct clause_dptr *cp;

  for (cp = conjecture.first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
  printf("\n");
  fflush(stdout);
} 

void print_base_nos()
{    
  struct clause_dptr *cp;

  for (cp = semantic_base.first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
  printf("\n");
  fflush(stdout);
}

void print_ext_list(struct external_list *elist)
{
  struct clause_dptr *cp;

  for (cp = elist->first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
} /* print_ext_list() ------------------------------------------------*/ 

void debug_scan(char *event)

{
  int num_sets, maxiset;
  struct clause_dptr *cp;

  if (!develop_semantics) return;

  for (num_sets = 0; num_sets <= num_maxisets; num_sets++){
    for (cp = sos_scan[num_sets].first_cp; cp; cp = cp->next) {
      if (cp->c->num_sets != num_sets) 
	abort_scott("clause %d gets misplaced in sos_scan[]", cp->c->id);
    }
  }
  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
    for (cp = failed_clauses[maxiset].first_cp; cp; cp = cp->next) {
      if (cp->c->in_set[maxiset]) 
	abort_scott("clause %d gets misplaced in failed_clauses[]", cp->c->id);
    }
  }
  
} /*check_scan() -----------------------------------------------------*/


void print_sos()
{  
  struct clause *cl;

  printf("---->sos clauses\n");
  for (cl=Sos->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d id=", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }
} /* print_sos() -----------------------------------------------------*/

void print_base()
{    
  struct clause_dptr *cp;

  printf("---->base clauses\n");
  for (cp = semantic_base.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
}

void print_new_base()
{    
  struct clause_dptr *cp;

  printf("---->new base clauses\n");
  for (cp = new_base_clauses.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
}

void print_conjecture()
{    
  struct clause_dptr *cp;

  printf("---->conjecture clauses\n");
  for (cp = conjecture.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
} 

void print_usable()

{
  struct clause *cl;

  printf("---->usable clauses\n"); 
  for (cl=Usable->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d id=", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }

} /* print_usable() --------------------------------------------------*/

void print_first()

{
  struct clause *cl;

  printf("---->input_first clauses\n"); 
  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d ", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }

} /* print_first() ---------------------------------------------------*/

