/*--------------------------------------------------------------------
 *                        Scontrol.c                                  *
 *                                                                    *
 *       SCOTT functions that control the size and composition of the *
 *       clause space on the clause space                             *
 *                                                                    *
 *--------------------------------------------------------------------*/

#include "SCOTT.h"

static struct Distn pick_weight;
static struct Distn max_vars;

int clause_pick_weight(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by pick_weight->property()
   * Returns the pick weight of the clause
   *-----------------------------------------------------------------*/
{
  return cl->pick_weight;

}/*clause_pick_weight()----------------------------------------------*/

void init_distributions()
  /*------------------------------------------------------------------
   * Called by init().
   * sets up the distributions that we will monitor.
   * -----------------------------------------------------------------*/
{
  init_a_distribution(&pick_weight);
  init_a_distribution(&max_vars);

  pick_weight.name = "weight";
  max_vars.name =    "variable";

  pick_weight.property = clause_pick_weight;
  max_vars.property = distinct_vars;

  pick_weight.limit = &(Parms[MAX_WEIGHT].val);
  max_vars.limit = &(Parms[MAX_DISTINCT_VARS].val);

  pick_weight.cull_over = &(tParms[CULL_WEIGHTS_OVER].val);
  max_vars.cull_over = &(tParms[CULL_VARS_OVER].val);

  pick_weight.limit_at = &(tParms[LIMIT_WEIGHT_AT].val);
  max_vars.limit_at = &(tParms[LIMIT_VARS_AT].val);
  
  pick_weight.interval = max_vars.interval = -1;

}/*init_distributions()-----------------------------------------------*/

void init_a_distribution(struct Distn *distn)
  /*------------------------------------------------------------------
   * Called by init_distributions()
   * Initialise fields in the given Distn structure.
   * -----------------------------------------------------------------*/
{
 int class, maxiset;

  // re-init distributions
  for (class = 0; class < MAX_DIST; class++) {
    distn->freq[class] = 0;
    for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	distn->freq_over_set[maxiset][class] = 0;
  }
}/*init_a_distribution()----------------------------------------------*/

void get_distn(struct Distn *distn)
  /*------------------------------------------------------------------
   * Called by set_distn()
   *
   * Scans the SOS compiling frequency estimates for "distn".  Assumes
   * that the max_wt and min_wt fileds of distn have been set to
   * sensible values.
   * Assumes:
   *   MAX_DIST > 0
   *   distn->max != distn->min
   * ------------------------------------------------------------------*/
{
  struct clause *cl;
  int class, maxiset, diff, sum;
  float median_mark, average;
  int interval, max, min;

  init_a_distribution(distn); // reset frequency values

  max = distn->max;
  min = distn->min;
  average = distn->average;

  // select an appropriate interval length
  diff = max - min;
  if (diff % MAX_DIST) interval = (diff/MAX_DIST) + 1;
  else interval = diff/MAX_DIST;
  
  distn->interval = interval;

  // scan for frequencies in corresponding ranges
  sum = 0;
  for (cl = Sos->first_cl; cl; cl = cl->next_cl) {
    diff = distn->property(cl) - average;
    sum += diff * diff;
    class = ((distn->property(cl) - min) / interval);

    // skip outliers
    if (class < 0 || class >= MAX_DIST) continue;

    distn->freq[class]++;
    if (Flags[SEMANTIC_GUIDANCE].val) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++){
	if (active[maxiset] && cl->in_set[maxiset]) 
	  distn->freq_over_set[maxiset][class]++;
      }
    }
  }
  distn->variance = sum/(float)Stats[SOS_SIZE];
  
  // estimate the median
  median_mark = (Stats[SOS_SIZE]+1)/2;
  class = 0;
  while (class < MAX_DIST) {
    if (median_mark < distn->freq[class]) break;
    median_mark -= distn->freq[class];
    class++;
  }
  if (!distn->freq[class]) {
    distn->median = (class*interval)+min;
  }
  else {
    distn->median = (class*interval)+min+
      (interval*(median_mark/distn->freq[class]));
  }
}

void set_distn(struct Distn *distn)
  /*------------------------------------------------------------------*
   * Called by percentile()
   * 
   * Scans SOS for max/min values and sets so outliers are excluded,
   * then calls get_distn() to compile frequency estimates for
   * "distn".
   * ------------------------------------------------------------------*/
{
  struct clause *cl;
  int class, max_freq;

  if (Stats[SOS_SIZE] == 0) return;

  distn->min = MAX_INT;
  distn->max = 0;
  distn->interval = 1;

  if (Stats[SOS_SIZE] == 0){
    message(0,"WARNING: set_distn() called with an empty SOS");
    return;
  }
  // scan for min and max value of the distribution
  for (cl = Sos->first_cl; cl; cl = cl->next_cl) {
    if (distn->property(cl) > distn->max) distn->max =  distn->property(cl);
    if (distn->property(cl) < distn->min) distn->min =  distn->property(cl);
  }

  if (distn->max == distn->min) {
    distn->average = distn->median = distn->max;
    distn->freq[0] = Stats[SOS_SIZE];
    return;
  }
  distn->average = (float)(Stats[SOS_WT]/Stats[SOS_SIZE]);  
  
  get_distn(distn);

  // check for outliers
  for (class = 0, max_freq = 0; class < MAX_DIST; class++)
    if (distn->freq[class] > max_freq) max_freq = distn->freq[class];

  // first upwards outliers
  distn->max_outlier = distn->max;
  distn->num_outliers_up = 0;
  for (class = MAX_DIST-1; class >= 0; class--){
    if (distn->freq[class] > max_freq/(float)(100)) {
      distn->max = ((class+1)*(distn->interval))+(distn->min)-1;
      break;
    }
    distn->num_outliers_up += distn->freq[class];
  }
 
  // next down wards outliers
  distn->min_outlier = distn->min;
  distn->num_outliers_down = 0;
  for (class = 0; class < MAX_DIST; class++){
    if (distn->freq[class] > max_freq/(float)(100)) {
      distn->min = (class*(distn->interval))+(distn->min);
      break;
    }
    distn->num_outliers_down += distn->freq[class];
  }
  
  // if max or min have changed we recompile freqencies
  if (distn->num_outliers_up || distn->num_outliers_down) {
    if (distn->max == distn->min) {
      distn->average = distn->median = distn->max;
      distn->freq[0] = Stats[SOS_SIZE];
      return;
    }
    get_distn(distn);
  }
}/*set_distn()--------------------------------------------------------*/


void output_distn(struct Distn *distn)
  /*------------------------------------------------------------------
   * Called by OTTER fcn main().  
   * 
   * Outputs formated freqency counts (and stats) for "distn".
   * Assumes distn has been set.
   * -----------------------------------------------------------------*/
{
  int class, maxiset;

  if (Stats[SOS_SIZE] == 0) return;

  printf("\n-------------- Begin %s distibution ---------------\n", distn->name);
  printf("\nmin %d, max %d, average %.1f, median %.1f, variance %.1f"
	 "\n\n", 
	 distn->min, 
	 distn->max, 
	 distn->average, 
	 distn->median, 
	 distn->variance);
  
  if (distn->interval > 1) printf("  interval    :  freq");
  else  printf("value :  freq");
  if (Flags[SEMANTIC_GUIDANCE].val) {
    for (maxiset = 0; maxiset <=last_set_index; maxiset++)
      if (active[maxiset]) printf(" %5d", maxiset);
  }
  printf("\n");

  if (distn->interval > 1) printf("---------------------");
  else  printf("-------------");
  if (Flags[SEMANTIC_GUIDANCE].val) {
    for (maxiset = 0; maxiset <=last_set_index; maxiset++)
      if (active[maxiset]) printf("------");
  }
  printf("\n");
  if (distn->num_outliers_down) {
    if (distn->interval > 1) {
      printf("%5d - %-5d : %5d\n",
	     distn->min_outlier, distn->min - 1, 
	     distn->num_outliers_down);
    }
    else {
      printf("%5d : %5d\n",
	     distn->min_outlier,
	     distn->num_outliers_down);
    }
    printf("\n");
  }
  // print the frequencies
  for (class = 0; class < MAX_DIST; class++) {
    if ((class*distn->interval)+distn->min > distn->max) break;

    if (distn->interval > 1) {
      printf("%5d - %-5d : %5d",
	     (class*distn->interval)+distn->min,
	     ((class+1)*distn->interval)+distn->min-1,
	     distn->freq[class]);
    }
    else  {
      printf("%5d : %5d",
	     class+distn->min,
	     distn->freq[class]);
    }

    if (Flags[SEMANTIC_GUIDANCE].val) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) 
	  printf(" %5d",distn->freq_over_set[maxiset][class]);
    }
    printf("\n");
  }
  if (distn->num_outliers_up) {
    printf("\n");
    if (distn->interval > 1) {
      printf("%5d - %-5d : %5d\n",
	     distn->max+1,
	     distn->max_outlier, 
	     distn->num_outliers_up);
    }
    else {
      printf("%5d : %5d\n",
	     distn->max_outlier,
	     distn->num_outliers_up);
    }
  }
  printf("\n-------------- End %s distribution -----------------\n",distn->name);
  fflush(stdout); 
  
}/*--output_distn()---------------------------------------------------*/


void output_weight_distn()
  /*------------------------------------------------------------------
   * Called by main() Prints a leaf plot of the current pick weight
   * frequency distribution.
   * -----------------------------------------------------------------*/
{
  if (pick_weight.interval < 0)   set_distn(&pick_weight);
  message(2,"");
  output_distn(&pick_weight);
  message(2,"");

}/*output_weight_distn()--------------------------------------------------*/

void output_variable_distn()
  /*------------------------------------------------------------------
   * Called by main() Prints a leaf plot of the current pick weight
   * frequency distribution.
   * -----------------------------------------------------------------*/
{
  if (max_vars.interval < 0)   set_distn(&max_vars);
  message(2,"");
  output_distn(&max_vars);
  message(2,"");

}/*output_variable_distn()--------------------------------------------------*/

int percentile(struct Distn *distn, int percent)
  /*------------------------------------------------------------------
   * Called by start_scan() and control_memory().
   * 
   * Returns the maximum value of the given percentile in the given
   * "distn". Assumes the frequency distribution has been set.
   * -----------------------------------------------------------------*/
{
  int class, class_in_set, maxiset;
  float percentile_mark, interpolant;
  int value;
  char str[10];
  
  // estimate the top value in the percentile 
  percentile_mark = (Stats[SOS_SIZE]*percent)/(float)100;
  class = 0;
  while (class < MAX_DIST) {
    if (percentile_mark < distn->freq[class]) break;
    percentile_mark -= distn->freq[class];
    class++;
  }

  if (distn->freq[class]) 
    interpolant = 
      distn->interval
      *percentile_mark
      /(float)distn->freq[class];
  else 
    interpolant = 0;


  if (Flags[SEMANTIC_GUIDANCE].val) {
    // check percentile across maxisets
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (!active[maxiset]) continue;
      percentile_mark = (get_size_on_sos(maxiset)*percent)/(float)100;

      // skip empty theories
      if (percentile_mark == 0) continue;

      class_in_set = 0;
      while (class_in_set < MAX_DIST) {
	if (percentile_mark 
	    < distn->freq_over_set[maxiset][class_in_set]) break;
	percentile_mark -= distn->freq_over_set[maxiset][class_in_set];
	class_in_set++;
      } 
      if (class_in_set > class) {//chose the higher value
	class = class_in_set;
	if (distn->freq_over_set[maxiset][class])
	  interpolant = 
	    distn->interval
	    *percentile_mark
	    /(float)distn->freq_over_set[maxiset][class];
	else
	  interpolant = 0;
      }
    }
  }
  
  // kludge rounding rather than use a math library
  sprintf(str,"%.0f",
	  (class*distn->interval)+distn->min+interpolant);
  sscanf(str,"%d",&value);
  
  return value;
  
}/*--percentile()-----------------------------------------------------*/


void cull_clauses_over(int max_value, int (*property)())
  /*------------------------------------------------------------------
   * Called by start_scan().  Scan SOS discarding clause that are
   * heavier then the new max_weight. Must be called after the scan is
   * complete so that partition and scan data structures are correct.
   * -----------------------------------------------------------------*/
{  
  struct clause *cl, *next_cl;
  
  cl = Sos->first_cl;
  while (cl) {
    next_cl = cl->next_cl;
    if ((*property)(cl) > max_value){	
      CLOCK_START(UN_INDEX_TIME);
      un_index_lits_all(cl);
      CLOCK_STOP(UN_INDEX_TIME);
      rem_from_list(cl);
      hide_clause(cl);
    }
    cl = next_cl;
  }
  
}/*--cull_clauses_over()----------------------------------------------*/


int weight_percentile(int percent)
  /*------------------------------------------------------------------
   * Called by control_memory()
   *
   * Returns the maximum value of the given percentile in the
   * "pick_weight" distn.
   * -----------------------------------------------------------------*/
{
  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) 
    return MAX_INT;
  
  set_distn(&pick_weight);
  
  if ((pick_weight.max - pick_weight.min)*2 < MAX_DIST)
    return MAX_INT;

  return percentile(&pick_weight, percent);
  
}/*weight_percentile()------------------------------------------------*/

int variable_percentile(int percent)
  /*------------------------------------------------------------------
   * Called by control_memory() and start_scan().
   *
   * Returns the maximum value of the given percentile in the
   * "distinct_vars" distn.
   * -----------------------------------------------------------------*/
{
  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) 
    return MAX_INT;

  set_distn(&max_vars);

  if ((max_vars.max - max_vars.min)*2 < MAX_DIST)
    return MAX_INT;

  return percentile(&max_vars, percent);
  
}/*percentile()----------------------------------------------------*/



void scott_control_memory()
  /*------------------------------------------------------------------
   * Called by infer_and_process()
   * 
   * Checks memory allocation, and resets weight and variable limits
   * if appropriate.  This implements a soft limit on memory.
   * -----------------------------------------------------------------*/
{
  static int control_point = 0;
  int memory;

  /* should we control? */
  if (tParms[CONTROL_MEMORY_LIMIT].val == 0 ) return;
  memory = total_mem();
  if ((3*memory) < tParms[CONTROL_MEMORY_LIMIT].val) return;

  if (control_point && control_point != Stats[CL_GIVEN]) return;
  control_point = Stats[CL_GIVEN] + 20;
  control_distn(&pick_weight);
  control_distn(&max_vars);
}
/*control_memory()----------------------------------------------------*/


void control_distn(struct Distn *distn)
   /*------------------------------------------------------------------
    * Called by control_memory(). Examinies distribution and (possibly)
    * sets a new value for the associated limit.
    * -----------------------------------------------------------------*/
{
  int new_limit = MAX_INT;

  set_distn(distn);
  if ((distn->max - distn->min)*2 < MAX_DIST) return;

  new_limit = percentile(distn,*(distn->limit_at));

  if ((new_limit < *(distn->limit))
      || (*(distn->limit) == 0) ) {
    
    // set new value
    *(distn->limit) = new_limit;
    
    message(3,"");
    message(3,"Memory usage at control point %d Kbytes",total_mem());
    message(3,"");
    if (tParms[VERBOSITY].val > 2) output_distn(distn);
    message(2,"Resetting %s limit to %d "
	    "(search is now incomplete)\n",
	    distn->name, new_limit);
    message(2,"");
    
    new_limit = percentile(distn,*(distn->cull_over));
    cull_clauses_over(new_limit,distn->property);
  }
  
}/*--control_distn()-------------------------------------------------*/
