/*------------------------------------------------------------------------*
 *                         Sdeclare_symbols.c                             *
 *                                                                        *
 *  SCOTT functions for declaring OTTER symsbols to the FINDER module     *
 *                                                                        *
 *------------------------------------------------------------------------*/

#include "SCOTT.h"

#define MAX_SYMBOLS 100 // max symbols that we will bother modelling

static struct symbol_list {
  char name[MAX_NAME];
  int arity;
  int sym_num;
} 
Predicates[MAX_SYMBOLS], 
Functions[MAX_SYMBOLS], 
Constants[MAX_SYMBOLS];


boolean insert_predicate(int sym_num)
  /*------------------------------------------------------------------
   * Called by scan_list_for_symbols().  Inserts the "name" in the
   * symbol list for predicates iff it is not already there.
   * -----------------------------------------------------------------*/
{
  int entry;

  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    // break if symbol has already been defined
    if (Predicates[entry].sym_num == sym_num) return true;

    if (Predicates[entry].arity == -1) { // last entry
      strcpy(Predicates[entry].name,sn_to_str(sym_num));
      Predicates[entry].arity = sn_to_arity(sym_num);
      Predicates[entry].sym_num = sym_num;
      return true;
    }
  }

  message(1,"WARNING: MAX_SYMBOLS exceeded");
  return false;
  
} /*--insert_predicate() ---------------------------------------------*/


boolean insert_function(int sym_num)
  /*------------------------------------------------------------------
   * Called by scan_list_for_symbols().  Inserts the "name" in the
   * symbol list for functions iff it is not already there.
   * -----------------------------------------------------------------*/
{
  int entry;

  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    // break if symbol has already been defined
    if (Functions[entry].sym_num == sym_num) return true;

    if (Functions[entry].name[0] == '\0') { // last entry
      strcpy(Functions[entry].name,sn_to_str(sym_num));
      Functions[entry].arity = sn_to_arity(sym_num);
      Functions[entry].sym_num = sym_num;
      return true;
    }
  }

  message(1,"WARNING: MAX_SYMBOLS exceeded");
  return false;
  
} /*--insert_function() ----------------------------------------------*/


boolean insert_constant(int sym_num)
  /*------------------------------------------------------------------
   * Called by scan_list_for_symbols(). Inserts the "name" in the
   * symbol list for constants iff it is not already there.
   * -----------------------------------------------------------------*/
{
  int entry;

  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    // break if symbol has already been defined
    if (Constants[entry].sym_num == sym_num) return true;

    if (Constants[entry].name[0] == '\0') { // last entry
      strcpy(Constants[entry].name,sn_to_str(sym_num));
      Constants[entry].sym_num = sym_num;
      return true;
    }
  } 

  message(1,"WARNING: MAX_SYMBOLS exceeded");  
  return false;

} /*--insert_constant() ----------------------------------------------*/


boolean scan_term_for_symbols(struct term *subterm)
  /*------------------------------------------------------------------
   * Called by scan_list_for_symbols().  Scans the given subterm for
   * symbols.  Recursively calls itself.
   * -----------------------------------------------------------------*/
{
  struct rel *arg;

  if (subterm->type == NAME) {
    if (!insert_constant(subterm->sym_num)) return false;
  }
  else if (subterm->type == COMPLEX) {
    if (!insert_function(subterm->sym_num)) return false;
    for (arg = subterm->farg; arg; arg = arg->narg) 
      if (!scan_term_for_symbols(arg->argval)) return false;
  }

  return true;

} /*--scan_term_for_symbol()------------------------------------------*/


boolean scan_list_for_symbols(struct list *list_ptr)
  /*------------------------------------------------------------------
   * Called by scan_lists_for_symbols().  Scans the given clause list
   * for symbols.
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  struct literal *lit;
  struct rel *arg;

  for (cl = list_ptr->first_cl; cl; cl = cl->next_cl) {
    for (lit = cl->first_lit; lit; lit = lit->next_lit) {
      if (!insert_predicate(lit->atom->sym_num)) return false;

      if (lit->atom->type == COMPLEX) 
	for (arg = lit->atom->farg; arg; arg = arg->narg)
	  if (!scan_term_for_symbols(arg->argval)) return false;
    }
  }

  return true;

} /*--scan_list_for_symbols() ----------------------------------------*/



boolean scan_lists_for_symbols()
  /*------------------------------------------------------------------
   * Called by declare_otter_symbols().  Scans appropriate lists for
   * symbols.
   * -----------------------------------------------------------------*/
{

  if (scan_list_for_symbols(Usable) &&
      scan_list_for_symbols(Sos) &&
      scan_list_for_symbols(First) &&
      scan_list_for_symbols(Hot)) 
    return true;
  else 
    return false;

} /*--scan_lists_for_symbols()----------------------------------------*/

void debug_print_symbols()
{
  int entry;

  printf("\n\n%s","Constants: ");
  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    printf("%s:%d, ",Constants[entry].name,Constants[entry].arity);
  }

  printf("\n\n%s","Predicates: ");
  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    printf("%s:%d, ",Predicates[entry].name,Predicates[entry].arity);
  }

  printf("\n\n%s","Functions: ");
  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    printf("%s:%d, ",Functions[entry].name,Functions[entry].arity);
  }

  printf("\n");
  fflush(stdout);
}

boolean declare_otter_symbols()
  /*------------------------------------------------------------------
   * Called from init_finder().  
   * 
   * Ensures that any OTTER symbol can be interpreted by FINDER. This
   * is called after all OTTER input has been parsed and any
   * clausification processed.
   * -----------------------------------------------------------------*/
{
  int entry, arg;
  char declaration[CLAUSE_STRING_ARRAY];

  add_declaration("setting silent_breaks\n");
  add_declaration("setting verbosity { job: none. models: none. stats: none. }\n");

  sprintf(declaration,"sort default cardinality < %d\n",tParms[MAX_DOMAIN_SIZE].val+1);
  add_declaration(declaration);
  if (tParms[MIN_DOMAIN_SIZE].val > 0) 
    sprintf(declaration,"sort default cardinality > %d\n",tParms[MIN_DOMAIN_SIZE].val-1);
  add_declaration(declaration);

  // initialise
  for (entry = 0; entry < MAX_SYMBOLS; entry++){
    Predicates[entry].name[0] = '\0';
    Functions[entry].name[0] = '\0';
    Constants[entry].name[0] = '\0';
    Constants[entry].arity = -1;
    Predicates[entry].arity = -1;
    Functions[entry].arity = -1;
  }

  if (!scan_lists_for_symbols()) {
    return false;
  }
  //debug_print_symbols();

  printf("\n------------> declare semantics\n\n");
  strcpy(declaration,"");
  for (entry = 0; Constants[entry].name[0] != '\0'; entry++){
    sprintf(declaration,"constant %s default.",
	    Constants[entry].name);
    printf("%s\n",declaration);
    add_declaration(declaration);
  }

  strcpy(declaration,"");
  for (entry = 0; Predicates[entry].name[0] != '\0'; entry++){
    if (Predicates[entry].arity == 0) {
      sprintf(declaration,"constant %s bool.",Predicates[entry].name);
      printf("%s\n",declaration);
      add_declaration(declaration);
      continue;
    }
    sprintf(declaration,"function %s ",Predicates[entry].name);
    for (arg = 1; arg < Predicates[entry].arity; arg++){
      strcat(declaration, "default, ");
    }
    strcat(declaration, "default -> bool.");
    printf("%s\n",declaration);
    add_declaration(declaration);
  }

  strcpy(declaration,"");
  for (entry = 0; Functions[entry].name[0] != '\0'; entry++){
    sprintf(declaration,"function %s ",Functions[entry].name);
    for (arg = 1; arg < Functions[entry].arity; arg++){
      strcat(declaration, "default, ");
    }
    strcat(declaration, "default -> default {no_priority}");
    printf("%s\n",declaration);
    add_declaration(declaration);
  }

  if (equality_is_present()) {
    if (Flags[PROLOG_STYLE_VARIABLES].val)
      strcpy(declaration,"clause { equal(A,B) = (A = B). }");
    else // use (hopefuly) unique variable names
      strcpy(declaration,"clause { equal(sCoTtA,sCoTtB) = (sCoTtA = sCoTtB). }");
    printf("%s\n",declaration);
    add_declaration(declaration);
  }

  fflush(stdout);

  return true;
} /*--declare_otter_symbols()-----------------------------------------*/

