/*------------------------------------------------------------------------*
 *                         Spartition.c                                   *
 *                                                                        *
 *  SCOTT functions for maintaining partition in formation.               *
 *                                                                        *
 *------------------------------------------------------------------------*/

#include "SCOTT.h"
#include <stdio.h>
#include <stdarg.h>

static struct partition set_in_cell;
static struct partition cmp_in_cell;

/*--------------------------------------------------------------------
 * The following group of functions all return the value of a variable
 * that would otherwise be out of scope.
 * -------------------------------------------------------------------*/

int get_sos_partition(int maxiset)  { return set_in_cell.sos[0][maxiset];    }
int get_size_on_sos(int maxiset)    { return set_in_cell.sos[maxiset][0];    }
int get_size_on_usable(int maxiset) { return set_in_cell.usable[maxiset][0]; }


/*--------------------------------------------------------------------
 * The remaining functions in this file operate extensively on this
 * partition information.
 * -------------------------------------------------------------------*/


void print_a_partition(struct partition *part_ptr)
  /*------------------------------------------------------------------
   * Called by various print_partition().  Prints out the given
   * partition data appropriately formated.
   * -----------------------------------------------------------------*/
{
  int maxiset, cell;
  char str[15];
  
  // print USABLE partition
  printf("\n U ");
  for (cell = 0; cell <= num_maxisets; cell++) 
    printf("%5d",part_ptr->usable[0][cell]);
  printf(" | %5ld\n", Stats[USABLE_SIZE]);
  
  printf("------------");
  for (cell = 0; cell <= num_maxisets; cell++) printf("-----");
  printf("\n");
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      printf("%2d:     ",maxiset);
      for (cell = 1; cell <= num_maxisets; cell++)
	printf("%5d",part_ptr->usable[maxiset][cell]);
      printf(" | %5d\n",part_ptr->usable[maxiset][0]);
    }
  }

  // print SOS partition
  printf("\n S "); 
  for (cell = 0; cell <= num_maxisets; cell++) 
    printf("%5d",part_ptr->sos[0][cell]);
  printf(" | %5ld     rating\n",Stats[SOS_SIZE]);

  printf("------------");
  for (cell = 0; cell <= num_maxisets; cell++) printf("-----");
  printf("\n");
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      printf("%2d:     ",maxiset);
      for (cell = 1; cell <= num_maxisets; cell++)
	printf("%5d",part_ptr->sos[maxiset][cell]);
      //printf(" | %5d\n",part_ptr->sos[maxiset][0]);

      sprintf(str,"%+.1f",rate_maxiset(maxiset));
      printf(" | %5d     %6s\n",part_ptr->sos[maxiset][0],str);
    }
  }
  
  fflush(stdout);

} /*--print_a_partition()---------------------------------------------*/

void print_partition(char *event)   
  /*------------------------------------------------------------------
   * Called by various functions 
   * 
   * Prints a table describing the current state of the partition. The
   * string "event" indicates what imediately preceeded the call.
   * This also doubles as a usefull debugging tool.
   * -----------------------------------------------------------------*/
{
  if (tParms[VERBOSITY].val <= 1) return;

  if (tParms[VERBOSITY].val >= 4){
    printf("\nMaxiset partititon %s\n", event); 
    print_a_partition(&set_in_cell);
  }
  
  printf("\nComplement partititon %s\n", event); 
  print_a_partition(&cmp_in_cell);
  printf("\n");
  message(2,"Average size of choice space: %0.1f", size_choice_space());
  message(2,"");
  
} /*--print_partition()-----------------------------------------------*/


boolean subset(int set1, int set2)
  /*------------------------------------------------------------------
   * Called by already_exists() and discard_subsets().
   * 
   * Detemines whether "set1" is a (weak) subset of "set2".
   * -----------------------------------------------------------------*/
{
  int cell;
  struct clause *cl;

  // compare partitions
  for (cell=0;cell <= last_set_index; cell++){
    if (set_in_cell.sos[set1][cell] > 
	set_in_cell.sos[set2][cell]) return false; 
    if (set_in_cell.usable[set1][cell] > 
	set_in_cell.usable[set2][cell]) return false; 
  }

  // search for a cl in set1 that is not in set2 
  for (cl = Sos->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl =First->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl = Usable->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  
  return true;

} /* subset() --------------------------------------------------------*/

boolean sos_subset(int set1, int set2)
  /*------------------------------------------------------------------
   * Called by already_exists() and discard_subsets().  
   * 
   * Detemines whether "set1" is a (weak) subset of "set2", but
   * checking for disparity only on the SOS or INPUT_FIRST list.
   * -----------------------------------------------------------------*/
{
  int cell;
  struct clause *cl;

  // compare partitions
  for (cell=0;cell <= last_set_index; cell++){
    if (set_in_cell.sos[set1][cell] > 
	set_in_cell.sos[set2][cell]) return false; 
  }

  // search for a cl in set1 that is not in set2 
  for (cl =Sos->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl =First->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  
  return true;

} /* sos_subset() ---------------------------------------------------*/

boolean already_exists(int new_maxiset) 
  /*------------------------------------------------------------------
   * Called by keep().
   * 
   * Returns TRUE iff new_maxiset is a subset of an existing maxiset
   * -- this property is tested with respect to the SOS only.  Note:
   * returns FALSE if the maxiset contains no SOS clauses (and hence
   * is an SOS subset of all other maxisets).
   * -----------------------------------------------------------------*/
{
  int old_maxiset;

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] && old_maxiset !=  new_maxiset) {
      if (subset(new_maxiset,old_maxiset) // real_subset
	  //if (sos_subset(new_maxiset,old_maxiset) &&
	  && get_size_on_sos(old_maxiset) < Stats[SOS_SIZE]) {
	
	message(2,"Maxiset %d is a subset of the maxiset %d", 
		new_maxiset, old_maxiset);
	return true;
      }
    }
  }
  
  return false;
  
} /* already_exists() ------------------------------------------------*/

int superset_of(int new_maxiset) 
  /*------------------------------------------------------------------
   * Called by keep().
   * 
   * Returns the index of any maxiset that is a superset of the
   * new_maxiset.  Returns 0 if there is no such maxiset.  Since
   * Usable list is not rigourously tested, set membership may be
   * excessively inacurrate, hence we only check the SOS and
   * INPUT_FIRST lists.
   * -----------------------------------------------------------------*/
{
  int old_maxiset;

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] && old_maxiset !=  new_maxiset) {
      if (sos_subset(new_maxiset,old_maxiset)
	  && get_size_on_sos(old_maxiset) < Stats[SOS_SIZE]) {
	return old_maxiset;
      }
    }
  }
  
  return 0;
  
} /* superset_of() ---------------------------------------------------*/

void check_for_model(int new_maxiset)
  /*------------------------------------------------------------------
   * Called by execute_strategy().  Checks if a model of the whole
   * clause space has been generated.c
   * Either returns TRUE or does not return at all!!
   * -----------------------------------------------------------------*/
{
  // check if we have a model of the whole clause space
  if (set_in_cell.sos[new_maxiset][0] == Stats[SOS_SIZE] &&
      set_in_cell.usable[new_maxiset][0] == Stats[USABLE_SIZE]) {
    
    // this may be because of weak equality axioms
    if (equality_is_present() && 
	!model("equal(A,B) = (A = B).",new_maxiset, false)){
      // stick it in the base maxiset

      record_event(STRENGTHEN_EQ_DEF);
      
      message(1,"Strengthening equality definition");
      model("equal(A,B) = (A = B).",0,false);
      return;
    }
    verify_model(new_maxiset);    // may not return
    if (!Flags[SEMANTIC_GUIDANCE].val) return;
    output_the_model(new_maxiset);
    cleanup();
    exit(POSSIBLE_MODEL_EXIT);
  }
  
}/*--check_for_model()------------------------------------------------*/

float size_choice_space()
  /*------------------------------------------------------------------
   * Called by rate_partition().
   * 
   * Returns the average size of the current choice space.  This
   * costing provides a measure of the quality of choice: smaller
   * ratings are better.
   * -----------------------------------------------------------------*/
{
  int maxiset, whole;

  /* 
     average choice space of whole partition

       |C1|+|C2|+...+|Cn|+|M1^M2^...^Mn|
     = ---------------------------------
                     n+1
		    
     = whole/(n+1)

     Now |M1^M2^...^Mn| = |{cl:cl is in "n" maxisets}|
  */

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  return (whole/(float)(num_maxisets+1));
  
} /*--size_choice_space()---------------------------------------------*/

float rate_partition()
  /*------------------------------------------------------------------
   * Called by keep() and refine_choice().
   * 
   * Returns the ratio of the choice space size with semantic
   * selection versus the choice space size with lightest first
   * selection.
   * -----------------------------------------------------------------*/
{
  return (size_choice_space()/(float)Stats[SOS_SIZE]);

}/*rate_partition()---------------------------------------------------*/

float rate_maxiset(int skip_set)
  /*------------------------------------------------------------------
   * Called by rate_clause() and keep().
   * 
   * Returns the effect that keep "skip_set" has on the partition
   * rating.  This costing provides a measure of the quality of
   * choice: (larger) negative ratings are better because the maxiset
   * decreases the size choice space most.
   * -----------------------------------------------------------------*/
{
  int maxiset;
  float whole, part = 0;
  float rating_with_skip_set, rating_without_skip_set;

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  part = whole - cmp_in_cell.sos[skip_set][0] 
    + set_in_cell.sos[skip_set][num_maxisets-1];
  
  rating_with_skip_set = (whole/(float)(num_maxisets+1));
  rating_without_skip_set = (part/(float)(num_maxisets));
  
  return (rating_with_skip_set - rating_without_skip_set);
  
} /*--rate_maxiset()--------------------------------------------------*/


float rate_clause(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by extend_base().
   *
   * Returns the expected change in the size of the choice space if
   * all the maxisets that the given clause does not belong to are
   * discarded.
   *  
   * Since we dont back cull in this way any more the rating is
   * meaningless
   * 
   * This costing provides a measure of the quality of choice.
   * -----------------------------------------------------------------*/
{
  int maxiset, cmp, num_cmps;
  float whole, part = 0;
  float rating_with_skip_sets, rating_without_skip_sets;
  
  if (num_maxisets < 2) 
    abort_scott("rate_clause() called when num_maxisets < 2");

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  part = whole;
  num_cmps = num_maxisets - cl->num_sets;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++){
    if (active[maxiset] && !cl->in_set[maxiset]) {
      whole -= cmp_in_cell.sos[maxiset][0];
      for (cmp = 1; cmp <= num_cmps; cmp++) 
	whole += cmp_in_cell.sos[maxiset][num_maxisets-cmp];
    }
  }
  
  rating_with_skip_sets = (whole/(float)(num_maxisets+1));
  rating_without_skip_sets = (part/(float)(num_maxisets));
  
  return (rating_with_skip_sets - rating_without_skip_sets);
  
} /*--rate_clause()--------------------------------------------------*/


/*--------------------------------------------------------------------
 * The following functions specify the principle semantic startegies
 * -------------------------------------------------------------------*/

void set_sem_queue(struct list *lptr)
  /*------------------------------------------------------------------
   * Called by semantic_selection() iff SCOTT has been defined.
   * 
   * If performing a semantic queue, moves appropriate clauses to the
   * "input_fist" list. This function assumes the INPUT_FIRST list is
   * empty before it is called
   * -----------------------------------------------------------------*/
{
  struct clause *cl, *next_cl;
  int limit, maxiset,count;
  boolean first = true;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) {
    // move all SOS to INPUT_FIRST
    record_event(SELECT_QUEUE);
    cl = lptr->first_cl;
    while (cl) {
      next_cl = cl->next_cl;
      rem_from_list(cl);
      append_cl(First, cl);
      cl = next_cl;
      tStats[CL_SEM_QUEUED]++;
    }
    message(2,"Entire SOS placed in the semantic queue");
    return;
  }
  
  count = 0;
  limit = tParms[SEM_QUEUE_LIMIT].val;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {

    if (maxiset == last_queue_theory) continue;
    if (!active[maxiset] ) continue; 
    if (cmp_in_cell.sos[maxiset][1] == 0) continue; 
    if (cmp_in_cell.sos[maxiset][1] > limit) continue; 
    if (cmp_in_cell.usable[maxiset][1] > limit) continue; 

    count++;
    
    cl = lptr->first_cl;
    while (cl) {
      next_cl = cl->next_cl;
      if (!cl->in_set[maxiset]   // clause is in the complement
	  && cl->num_sets == num_maxisets-1 // only one complement
	  ) { 
	rem_from_list(cl);
	append_cl(First, cl);    // move to the input_first list
	tStats[CL_SEM_QUEUED]++;
	last_queue_theory = maxiset; // break cycles
	if (first) {
	  message(2,"");
	  message(2,"Following clauses placed in the semantic queue");
	  first = false;
	}
	if (tParms[VERBOSITY].val > 1) print_clause(stdout,cl);
      }
      cl = next_cl;
    }
  }
  if (count > 1) last_queue_theory = 0;

  if (!first) message(2,"");


}/*--set_sem_queue()----------------------------------------------*/

int next_set_to_skip()
  /*------------------------------------------------------------------
   * Called by find_best_cl().  
   * 
   * The next given clause will belong to the complement of the next
   * skip_set.  If there is more than one maxiset, then skip set is
   * set to the next highest index, cycling through the availiable
   * theories, and skipping empty theories.  If no clear skip set can
   * be established -1 is returned.  (This may never actually happen)
   * -----------------------------------------------------------------*/
{
  int skip_set, count;
  static int last_skip_set = -1;
  
  skip_set = last_skip_set;
  for (count = 0; count <= last_set_index; count++) {
    
    if (skip_set == last_set_index) skip_set = 0;
    else skip_set++;
    
    if (skip_set == last_skip_set) break; 
    
    if (active[skip_set] 
	// &&
	//	size_on_sos[skip_set] > 0 &&
	//      size_on_sos[skip_set] < Stats[SOS_SIZE]
	// ...  these guys covered by previous call to discard_bad_maxisets()
	){
      last_skip_set = skip_set;
      
      return skip_set;
    }
  }
  
  last_skip_set = -1;
  return -1;

}/* next_skip_set() --------------------------------------------------*/

int variance(struct clause *cl)
  /*-------------------------------------------------------------------*\
   * Called by find_best_sem_cl().  
   * 
   * Return the resulting variance of the distribution of complements
   * on the USABLE list that would be expected if cl were introduced.
   * Note: clauses with no models belong to every complement.
   * -----------------------------------------------------------------*/
{
  int maxiset;
  float diff = 0;
  float average = 0;
  float variance = 0;
  
  // get the new distribution
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      average += (cmp_in_cell.usable[maxiset][0]
		  +(int)(!cl->in_set[maxiset]));
    }
  }
  average /= num_maxisets;
  
  // get the new variance after cl is chosen
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      diff = (cmp_in_cell.usable[maxiset][0] 
	      + (int)(!cl->in_set[maxiset]) - average);
      // probably fewer operations this way
      variance += diff*diff;
    }
  }
  
  return (variance/num_maxisets);
  
}/*--variance() ------------------------------------------------------*/


boolean cover_new_clauses(int key)
  /*------------------------------------------------------------------*\
   * Called by execute_strategy().  
   *
   * Attempts to provide control so that the new maxisets will to
   * cover any SOS clauses that currently belong to none.
   * -----------------------------------------------------------------*/
{
  static int last_null_set_size = 0;
  
  switch(key) { 
    
  case RECORD_EVENT:

    record_event(COVER_NEW_CLAUSES);
    tStats[COVER_ATTEMPTS]++;
    message(2,"");
    message(2,"====> Attempting to cover new clauses");

    return true; // return value not important
    
  case TEST: // are there any new null clauses?

    if ( (set_in_cell.sos[0][0] - null_set_size) ) {
      last_null_set_size = set_in_cell.sos[0][0];
      return true;
    }
    return false;
    
  case FAILS: // have we failed to eliminated any null clauses?

    if (last_null_set_size <= null_set_size) {
      message(2,"Failed to cover any new clauses");
      
      inc_time_limit();
      if (null_set_size > tParms[MAX_NULL_SET_SIZE].val)
	inc_cardinality();
    }
    
    return false;
    
  case RECORD_KEPT:

    tStats[COVER_KEPTS]++;
    return true; // return value not important

  }
  
  abort_scott("cover_new_clauses() called with bad key %d.",key);
  
  return false; // to quiet -Wall
  
}/*--cover_new_clauses()----------------------------------------------*/

boolean refine_choice(int key)
  /*------------------------------------------------------------------*\
   * Called by execute_strategy().  
   *
   * Attempts to provide control so that the new maxisets will
   * decrease the size of the choice space.
   * -----------------------------------------------------------------*/
{
  static float last_rating = 0;
  float this_rating;
  
  switch(key) {   

  case RECORD_EVENT:

    record_event(REFINE_CHOICE);
    tStats[REFINE_ATTEMPTS]++;
    message(2,"");
    message(2,"====> Attempting to refine choice (%.2f)",last_rating);
    return true; // return value not important

  case TEST:  /* is the average size of the choice space more than
		 (adjusted) size of SOS*/

    last_rating = rate_partition(); 
    if (last_rating*fParms[GUIDANCE_FACTOR].val > 1)
      return true;
    else
      return false;

  case FAILS:    // previous rating was better?
    
    this_rating = rate_partition();
    if (last_rating > this_rating) return false;

    message(2,"Previous rating was better (%.2f <= %.2f)",
	    last_rating, this_rating);
    return true;

  case RECORD_KEPT:

    tStats[REFINE_KEPTS]++; 
    return true; // return value not important

  }
  
  abort_scott("refine_choice() called with bad key %d",key);
  
  return false; // to quiet -Wall
  
}/*--refine_choice()--------------------------------------------------*/


boolean in_key_set(int key_set, struct clause *cl)
{
  if (num_maxisets > 0 && key_set)
    return (cl->in_set[key_set]);
  else
    return false;

}/*in_key_set()-------------------------------------------------------*/


struct clause *search_list_1(struct list *lptr, int skip_maxiset)
  /*------------------------------------------------------------------*\
   * Called by extend_base().  
   *
   * Search given list for a clause with the following conditions: 
   * a). clause is not already be in the base, new base, or conjecture 
   * c). clause does not belong to the skip_maxiset (if given), 
   * b). clause is consistent with the base,
   * d). and clause does not belong to every maxiset,
   * e). preference is given to clauses with fewer maxiset.
   *
   * We search backwards so that new clauses get a chance to
   * participate.
   * ------------------------------------------------------------------*/
{
  struct clause *cl, *new_cl, *new_base_first;
  
  if (new_base_clauses.first_cp)
    new_base_first = new_base_clauses.first_cp->c;
  else new_base_first = NULL;

  for (new_cl = NULL, cl = lptr->last_cl; cl; cl = cl->prev_cl) {

    // skip base, conjecture, and clauses alrerady in the new base
    if (cl->in_base) continue; 
    // skip the key set
    // skip the key set
    if (num_maxisets > 0 &&  
	skip_maxiset && 
	cl->in_set[skip_maxiset]) 
      continue;
    
    if (num_maxisets > 1) {
      if (cl->num_sets == 0) continue; // is inconsistent with base
      if (cl->num_sets == num_maxisets) continue;//tautology?
      if (new_base_first 
	  && in_set_identical(cl,new_base_first))
	continue;
    }

    if (!new_cl	
	// || (cl->num_sets < new_cl->num_sets)
	|| (cl->num_sets > new_cl->num_sets)
	) new_cl = cl;
  }
  return new_cl;

}/*search_list_1()------------------------------------------------------*/

struct clause *search_list_2(struct list *lptr, int skip_maxiset)
  /*------------------------------------------------------------------*\
   * Called by extend_base().  As for search_list_1() but omitting the
   * condition that all suitable clauses dont have their in_base
   * marker set.
   * ------------------------------------------------------------------*/
{
  struct clause *cl, *new_cl, *new_base_first;

  if (new_base_clauses.first_cp)
    new_base_first = new_base_clauses.first_cp->c;
  else new_base_first = NULL;

  for (new_cl = NULL, cl = lptr->last_cl; cl; cl = cl->prev_cl) {

    // skip conjecture, and clauses already in the new base
    if (cl->in_base > 1) continue;

    // skip the key set
    if (num_maxisets > 0 &&  
	skip_maxiset && 
	cl->in_set[skip_maxiset]) 
      continue;

    if (num_maxisets > 1) {
      if (cl->num_sets == 0) continue; // is inconsistent with base
      if (cl->num_sets == num_maxisets) continue; //tautology?
      if (new_base_first 
	  && in_set_identical(cl,new_base_first))
	continue;
    }

    if (!new_cl	
	// || (cl->num_sets < new_cl->num_sets)
	|| (cl->num_sets > new_cl->num_sets)
	) new_cl = cl;
  }
  return new_cl;

}/*search_list_2()------------------------------------------------------*/

struct clause *search_list_3(struct list *lptr, int skip_maxiset)
  /*------------------------------------------------------------------*\
   * Called by extend_base().  As for search_list_2() but omitting the
   * condition that all suitable clauses belong to at least one
   * maxiset.
   * ------------------------------------------------------------------*/
{
  struct clause *cl, *new_cl;

  for (new_cl = NULL, cl = lptr->last_cl; cl; cl = cl->prev_cl) {

    // skip new base and conjecture clauses only
    if (cl->in_base > 1) continue;

    // skip the key set
    if (num_maxisets > 0 &&  
	skip_maxiset && 
	cl->in_set[skip_maxiset]) 
      continue;

    if (num_maxisets > 1) {
      if (cl->num_sets == num_maxisets) continue;
    }
    
    if (!new_cl
	//	|| (cl->num_sets < new_cl->num_sets)
	|| (cl->num_sets > new_cl->num_sets)
	) {
      new_cl = cl;
    } 
  }
  return new_cl;

}/*search_list_3()------------------------------------------------------*/


boolean model_usable()
  /*-----------------------------------------------------------------*\
   * Called by scan_usable().  Returns true iff USbale list contains
   * few hard to model clauses.
   * ---------------------------------------------------------------*/
{
  // this is just a stub at the moment
  if ((set_in_cell.usable[0][0] + set_in_cell.usable[0][1]) > 20)
    return false;
  else 
    return true;

}/*model_usable()----------------------------------------------------*/

/*--------------------------------------------------------------------
 * The following group of functions maintain the partition information
 * --------------------------------------------------------------------*/

void init_a_partition(struct partition *part_ptr)
  /*------------------------------------------------------------------
   * Called by init_semantics().  Initialise a partition structure.
   * -----------------------------------------------------------------*/
{
  int cell, maxiset = 0;

  // initialize the partition
  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++){
      part_ptr->sos[maxiset][cell]    = 0;
      part_ptr->usable[maxiset][cell] = 0;
    }
  }
  
}/*--init_a_partition() ----------------------------------------------*/

void init_partitions()
  /*------------------------------------------------------------------
   * Called by init_semantics().  Initialise a partition structure.
   * -----------------------------------------------------------------*/
{

  init_a_partition(&set_in_cell);
  set_in_cell.sos[0][0]    = Stats[SOS_SIZE];
  set_in_cell.usable[0][0] = Stats[USABLE_SIZE];

  init_a_partition(&cmp_in_cell);
  cmp_in_cell.sos[0][0]    = Stats[SOS_SIZE];
  cmp_in_cell.usable[0][0] = Stats[USABLE_SIZE];

}/*--init_partitions() ------------------------------------------------*/


void init_sizes(int maxiset)
  /*------------------------------------------------------------------
   * Called by SCOTT fcn discard(). Resets the sizes of the recently
   * discarded maxiset.
   * -----------------------------------------------------------------*/
{
  set_in_cell.sos[maxiset][0] = 0;
  cmp_in_cell.sos[maxiset][0] = 0;
  set_in_cell.usable[maxiset][0] = 0;
  cmp_in_cell.usable[maxiset][0] = 0;

}/* init_sizes() -----------------------------------------------------*/

void usable_part_insert_all(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by scott_insert().  Incorporates the new clause into in
   * the usable partition data structures.
   * -----------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.usable[0][num_sets]++;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]++;
	set_in_cell.usable[maxiset][0]++;
      }
    }
  }

  cmp_in_cell.usable[0][num_cmps]++;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]++;
	cmp_in_cell.usable[maxiset][0]++;
      }
    }
  }
} /*--usable_part_insert_all()----------------------------------------*/

void sos_part_insert_all(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by scott_insert() Incorporates the new clause into in the
   * sos partition data structures.
   * -----------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.sos[0][num_sets]++;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]++;
	set_in_cell.sos[maxiset][0]++;
      }
    }
  }
  
  cmp_in_cell.sos[0][num_cmps]++;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]++;
	cmp_in_cell.sos[maxiset][0]++;
      }
    }
  }
} /*--sos_part_insert_all()-------------------------------------------*/

void usable_part_delete_all(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by scott_delete(). Removes the old clause from the usable
   * partition data structures.
   * -----------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.usable[0][num_sets]--;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]--;
	set_in_cell.usable[maxiset][0]--;
      }
    }
  }
  
  cmp_in_cell.usable[0][num_cmps]--;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]--;
	cmp_in_cell.usable[maxiset][0]--;
      }
    }
  }
} /*--usable_part_delete_all()----------------------------------------*/

void sos_part_delete_all(struct clause *cl)
  /*------------------------------------------------------------------
   * Called by scott_delete(). Removes the old clause from the sos
   * partition data structures.
   * -----------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.sos[0][num_sets]--;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]--;
	set_in_cell.sos[maxiset][0]--;
      }
    }
  }
  
  cmp_in_cell.sos[0][num_cmps]--;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]--;
	cmp_in_cell.sos[maxiset][0]--;
      }
    }
  }
} /*--sos_part_delete_all()-------------------------------------------*/

void scott_delete_if_conjecture(struct clause *cl)
{
  if (cl->in_base == neither) {
    delete_cl_ext(cl,&conjecture);
    cl->in_base = false;
  }
}

void scott_insert_conjecture(struct clause *cl)
{
  if (cl->in_base != neither) {
    append_cl_ext(cl,&conjecture);
    cl->in_base = neither;
  }
}

void scott_insert(struct list *lptr, struct clause *cl)
  /*------------------------------------------------------------------
   * Called by append_cl() etc...  Incorporates the given clause and
   * the usable partition data structures after the addition of
   * new_maxiset.
   * -----------------------------------------------------------------*/
{
  /* New clauses have in_set[0] set to false; however we need to do
   * this if we are moving clauses about */
  if (cl->in_set[0]) null_set_size++;

  if (lptr == Usable) {
     if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[USABLE_WT] += cl->length;
      usable_part_insert_all(cl);
    }
    else Stats[USABLE_WT] += cl->pick_weight;
  }

  else if (lptr == Sos) { 
    if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[SOS_WT] += cl->length;
      if (develop_semantics) sos_scan_insert(cl);
      sos_part_insert_all(cl);
    }
    else Stats[SOS_WT] += cl->pick_weight;
  }
 
  else if (lptr == First) {
    Stats[SOS_SIZE]++;
    if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[SOS_WT] += cl->length;
      sos_part_insert_all(cl);
    }
    else Stats[SOS_WT] += cl->pick_weight;
  }  
  /*
  else if (lptr == Demodulators) {
    if (Flags[SEMANTIC_GUIDANCE].val) {
      scott_insert_conjecture(cl);
    }
  }
  */
} /*--scott_insert()--------------------------------------------------*/


void scott_delete(struct list *lptr, struct clause *cl) 
  /*------------------------------------------------------------------
   * Called by rem_from_list(). Incorporates the given clause and the
   * usable partition data structures after the addition of
   * new_maxiset.
   * -----------------------------------------------------------------*/
{
  if (cl->in_set[0]) null_set_size--;

  if (lptr == Usable) {
     if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[USABLE_WT] -= cl->length;
      usable_part_delete_all(cl);
    }
    else Stats[USABLE_WT] -= cl->pick_weight;
  }

  else if (lptr == Sos) { 
    if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[SOS_WT] -= cl->length;
      if (develop_semantics) sos_scan_delete(cl);
      sos_part_delete_all(cl);
    }
    else Stats[SOS_WT] -= cl->pick_weight;
  }
 
  else if (lptr == First) {
    Stats[SOS_SIZE]--;
    if (Flags[SEMANTIC_GUIDANCE].val) {
      Stats[SOS_WT] -= cl->length;
      sos_part_delete_all(cl);
    }
    else Stats[SOS_WT] -= cl->pick_weight;
  }
  /*
  else if (lptr == Demodulators) {
    if (Flags[SEMANTIC_GUIDANCE].val) {
      scott_delete_if_conjecture(cl);
    }
  }
  */

} /*--scott_delete()--------------------------------------------------*/


void usable_part_insert(struct clause *cl, int new_maxiset) 
  /*------------------------------------------------------------------
   * Called by update_partition().  Incorporates the given clause into
   * the new_maxiset in the usable partition data structures..
   * -----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[new_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    set_in_cell.usable[0][num_sets-1]--;
    set_in_cell.usable[0][num_sets]++;
    set_in_cell.usable[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]++;
	if (maxiset != new_maxiset)
	  set_in_cell.usable[maxiset][num_sets-1]--;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    cmp_in_cell.usable[0][num_cmps-1]--;
    cmp_in_cell.usable[0][num_cmps]++;
    cmp_in_cell.usable[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]++;
	if (maxiset != new_maxiset)
	  cmp_in_cell.usable[maxiset][num_cmps-1]--;
      }
    }
  }
} /* usable_part_insert()---------------------------------------------*/

void sos_part_insert(struct clause *cl, int new_maxiset) 
  /*------------------------------------------------------------------
   * Called by update_partition(). Incorporates the given clause into
   * the sos partition data structures after the addition of the new
   * maxiset.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[new_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    set_in_cell.sos[0][num_sets-1]--;
    set_in_cell.sos[0][num_sets]++;
    set_in_cell.sos[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]++;
	if (maxiset != new_maxiset)
	  set_in_cell.sos[maxiset][num_sets-1]--;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    cmp_in_cell.sos[0][num_cmps-1]--;
    cmp_in_cell.sos[0][num_cmps]++;
    cmp_in_cell.sos[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]++;
	if (maxiset != new_maxiset)
	  cmp_in_cell.sos[maxiset][num_cmps-1]--;
      }
    }
  }
} /* sos_part_insert() -----------------------------------------------*/


void usable_part_delete(struct clause *cl, int old_maxiset) 
  /*------------------------------------------------------------------
   * Called by reset_partition().  Removes the given clause from the
   * old_maxiset in the usable partition.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[old_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    
    set_in_cell.usable[0][num_sets]--;
    set_in_cell.usable[0][num_sets-1]++;
    //    set_in_cell.usable[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]--;
	if (maxiset != old_maxiset)
	  set_in_cell.usable[maxiset][num_sets-1]++;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    
    cmp_in_cell.usable[0][num_cmps]--;
    cmp_in_cell.usable[0][num_cmps-1]++;
    //    cmp_in_cell.usable[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]--;
	if (maxiset != old_maxiset)
	  cmp_in_cell.usable[maxiset][num_cmps-1]++;
      }
    }
  }
} /* usable_part_delete() --------------------------------------------*/


void sos_part_delete(struct clause *cl, int old_maxiset) 
  /*------------------------------------------------------------------
   * Called by reset_partition().  Removes the given clause from the
   * old_maxiset in the sos partition.
   * -----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[old_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    
    set_in_cell.sos[0][num_sets]--;
    set_in_cell.sos[0][num_sets-1]++;
    //    set_in_cell.sos[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]--;
	if (maxiset != old_maxiset)
	  set_in_cell.sos[maxiset][num_sets-1]++;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    
    cmp_in_cell.sos[0][num_cmps]--;
    cmp_in_cell.sos[0][num_cmps-1]++;
    //    cmp_in_cell.sos[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]--;
	if (maxiset != old_maxiset)
	  cmp_in_cell.sos[maxiset][num_cmps-1]++;
      }
    }
  }
} /* sos_part_delete() -----------------------------------------------*/


void sos_part_mv_to_set(struct clause *cl, int old_maxiset)
  /*------------------------------------------------------------------
   * Called by retest_failed_clauses().  Clause that was in the
   * complement of "old_maxiset" is moved to the "old_maxiset".
   * -----------------------------------------------------------------*/
{
  int num_cmps, num_sets, maxiset;
  
  num_sets = cl->num_sets;
  num_cmps = num_maxisets - cl->num_sets;

  // move to new set 
  set_in_cell.sos[0][num_sets-1]--;
  set_in_cell.sos[0][num_sets]++;
  set_in_cell.sos[old_maxiset][0]++;

  // remove from previous cmp
  cmp_in_cell.sos[0][num_cmps+1]--;
  cmp_in_cell.sos[0][num_cmps]++;
  cmp_in_cell.sos[old_maxiset][0]--;

  // was once in cmp of old_maxiset
  cmp_in_cell.sos[old_maxiset][num_cmps+1]--;

  // adjust table
  for (maxiset=1; maxiset <= last_set_index; maxiset++){
    if (!active[maxiset]) continue;

    if (cl->in_set[maxiset]) {
      if (num_sets>1 && maxiset != old_maxiset) 
	set_in_cell.sos[maxiset][num_sets-1]--;
      set_in_cell.sos[maxiset][num_sets]++;
    }
    else {
      cmp_in_cell.sos[maxiset][num_cmps+1]--;
      if (num_cmps>0) cmp_in_cell.sos[maxiset][num_cmps]++;
    }
  }
  
}/*--sos_part_mv_to_set()---------------------------------------------*/

void usable_part_mv_to_set(struct clause *cl, int old_maxiset)
  /*------------------------------------------------------------------
   * Called by retest_failed_clauses().  Clause that was in the
   * complement of "old_maxiset" is moved to the "old_maxiset".
   * -----------------------------------------------------------------*/
{
  int num_cmps, num_sets, maxiset;
  
  num_sets = cl->num_sets;
  num_cmps = num_maxisets - cl->num_sets;

  // move to new set 
  set_in_cell.usable[0][num_sets-1]--;
  set_in_cell.usable[0][num_sets]++;
  set_in_cell.usable[old_maxiset][0]++;

  // remove from previous cmp
  cmp_in_cell.usable[0][num_cmps+1]--;
  cmp_in_cell.usable[0][num_cmps]++;
  cmp_in_cell.usable[old_maxiset][0]--;

  // was once in cmp of old_maxiset
  cmp_in_cell.usable[old_maxiset][num_cmps+1]--;

  // adjust table
  for (maxiset=1; maxiset <= last_set_index; maxiset++){
    if (!active[maxiset]) continue;

    if (cl->in_set[maxiset]) {
      if (num_sets>1 && maxiset != old_maxiset) 
	set_in_cell.usable[maxiset][num_sets-1]--;
      set_in_cell.usable[maxiset][num_sets]++;
    }
    else {
      cmp_in_cell.usable[maxiset][num_cmps+1]--;
      if (num_cmps>0) cmp_in_cell.usable[maxiset][num_cmps]++;
    }
  }
  
}/*--usable_part_mv_to_set()---------------------------------------------*/

/*================== DEBUGGING TOOLS =================================*/


void init_partition_box(int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY])
{
  int cell, maxiset;

  // Initialize the list partition copy
  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      set_in_list_cell[maxiset][cell] = 0;
    }
  }
  
}/*--init_partition_box()---------------------------------------------*/

int get_list_part(char *list_name, struct list *lptr,
		  int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		  int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY])
  
{
  int maxiset, num_sets, num_cmps;
  struct clause *cl;
  int errors = 0;
  
  for (cl=lptr->first_cl;cl;cl=cl->next_cl) {
    
    num_sets = 0;
    for (maxiset = 1; maxiset <= last_set_index; maxiset++)
      if (active[maxiset] && cl->in_set[maxiset])
	num_sets++;
    
    if (cl->num_sets != num_sets) {
      printf("ERROR: Clause %d on %s gets wrong num_sets.\n",cl->id,list_name);
      errors++;
    }
    
    if (cl->pick_weight != (semantic_wt(cl))) {
      printf("ERROR: Clause %d on %s gets bad weight.\n",cl->id,list_name);
      errors++;
    }    

    num_cmps = num_maxisets - num_sets;

    set_in_list_cell[0][num_sets]++;
    cmp_in_list_cell[0][num_cmps]++;

    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset]) {
	if (cl->in_set[maxiset]) {
	  set_in_list_cell[maxiset][num_sets]++;
	  set_in_list_cell[maxiset][0]++;
	}
	else {
	  cmp_in_list_cell[maxiset][num_cmps]++;
	  cmp_in_list_cell[maxiset][0]++;
	}
      }
    }
  }

  return errors;

}/*--get_list_part()--------------------------------------------------*/


int compare_partitions(char *list_name,
		       int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int set_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY])

{
  int cell, maxiset;
  int errors = 0;

  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && 
	  set_in_list_cell_copy[maxiset][cell] != 
	  set_in_list_cell[maxiset][cell]) {
	printf("ERROR: set_in_cell.%s[%d][%d]) = %d but should be %d.\n",
	       list_name, 
	       maxiset,
	       cell,
	       set_in_list_cell[maxiset][cell],
	       set_in_list_cell_copy[maxiset][cell]);
	errors++;
      }
    }
  }

  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && 
	  cmp_in_list_cell_copy[maxiset][cell] != 
	  cmp_in_list_cell[maxiset][cell]) {
	printf("ERROR: cmp_in_cell.%s[%d][%d]) = %d but should be %d.\n",
	       list_name, 
	       maxiset,
	       cell,
	       cmp_in_list_cell[maxiset][cell],
	       cmp_in_list_cell_copy[maxiset][cell]);
	errors++;
      }
    }
  }

  
  return errors;

}/*--compare_partitions()---------------------------------------------*/


void debug_partition(char *event)
  /*------------------------------------------------------------------
   * Called by OTTER fcn main() and various other functions.
   * 
   * At end of search (while clause space is still intact) or while
   * debugging, checks that weights have been maintained properly and
   * the partition information is correct.
   * 
   * Exits iff a bad value has been detected.
   * -----------------------------------------------------------------*/
{
  int errors = 0;
  int count, maxiset;
  struct partition set_in_cell_cp;
  struct partition cmp_in_cell_cp;
  
  init_a_partition(&set_in_cell_cp);
  init_a_partition(&cmp_in_cell_cp);

  count = 0;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) count++;

  if (num_maxisets != count) {
    printf("ERROR: num_maxisets = %d but should be %d.\n",
	   num_maxisets, count);
    errors++;
  }

  // check the SOS partitions
  errors += get_list_part("sos", Sos, 
			  set_in_cell_cp.sos,
			  cmp_in_cell_cp.sos);

  errors += get_list_part("input_first", First, 
			  set_in_cell_cp.sos,
			  cmp_in_cell_cp.sos);
  
  errors += compare_partitions("sos",
			       set_in_cell.sos, set_in_cell_cp.sos,
			       cmp_in_cell.sos, cmp_in_cell_cp.sos);

  // check the USABLE partitions
  errors += get_list_part("usable", Usable, 
			  set_in_cell_cp.usable,
			  cmp_in_cell_cp.usable);

  errors += compare_partitions("usable",
			       set_in_cell.usable, set_in_cell_cp.usable,
			       cmp_in_cell.usable, cmp_in_cell_cp.usable);

  if (errors) {
    message(0,"");
    abort_scott("%d errors found in the partition record %s",
		errors,event);
  }
}/*--debug_partition()------------------------------------------------*/









